from .combo import combo
from .. import elements as el
from ..elements.prop_utils import to_rgba, rgba_to_hex
from ..elements.props import BoolProp, OptionProp, ColorProp, Prop, TypedListProp


def line_chart(*data, **props):
    return chart(*data, chart_type="line", **props)


def bar_chart(*data, **props):
    """This component renders bar charts. See [chart](/components/chart)."""
    return chart(*data, chart_type="bar", **props)


def scatter_chart(*data, **props):
    return chart(*data, chart_type="scatter", **props)


class AxisProp(OptionProp):
    """
    An `AxisProp` stores the configuration of an axis in a
    [chart](/components/chart).

    It takes as a value either a string, representing an axis type, or
    a list or tuple of axis labels.

    Here are the the possible axis types:

    * `"value"`. Represents a numeric axis. Values on this axis can be
      any number, and labels will be intelligently generated to fit
      the data.


    * `"category"`. Represents an axis made of discrete labels. For
      data points that do not provide a value for this axis, one will
      be automatically generated. The generated names start at "0" and
      increment by 1. The axis labels are rendered in the order
      encountered in the data (or automatically generated).

    * `"time"`. Represents a time axis. Values on this axis are time
      values, such as date strings or integers representing
      milliseconds since Epoch. Labels are intelligently generated to
      fit the data.

    * `"log"`. Like `"value"` but log data.

    Alternatively, a tuple or list is accepted, representing the total
    list of x-values. In this case, the axis type is implicitly set to
    `"category"`.

    Examples:
    ```py
    "value"
    "category"
    "log"
    "time"
    ("One", "Two", "Three")
    ```
    """

    def __init__(self, value, **kwargs):
        super().__init__(value, ("value", "time", "log", "category"), **kwargs)

    def parse(self, axis):
        try:
            value = super().parse(axis)
            return dict(type=value, splitLine=dict(show=False))
        except Exception:
            pass

        if axis is None:
            return dict(show=False)
        if type(axis) in (list, tuple):
            return dict(type="category", data=axis, splitLine=dict(show=False))
        raise ValueError("Invalid axis.")


class TooltipProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(value, (None, "axis", "item"), **kwargs)

    def parse(self, tooltip):
        value = super().parse(tooltip)
        if value is None:
            return None
        else:
            return dict(trigger=value)


def parse_series(chart_type, data, series_names):
    return [
        dict(
            type=chart_type,
            data=series,
            name=index if len(series_names) == 0 else series_names[index],
        )
        for index, series in enumerate(data)
    ]


def complementary_color(color):
    r, g, b, a = to_rgba(color)
    rgba = (255 - r, 255 - g, 255 - b, a)
    return rgba_to_hex(rgba)


class ChartDataProp(Prop):
    def parse(self, data):
        out_data = []
        has_pairs = False
        has_items = False

        for d in data:
            if len(d) == 0:
                out_data.append([])
                continue
            if isinstance(d, dict):
                if not all([isinstance(v, (int, str, float)) for v in d.values()]):
                    raise ValueError(
                        "Series specified with dict has non-primitive values."
                    )
                out_data.append(list(d.items()))
                has_pairs = True
            elif all([isinstance(x, (tuple, list)) and len(x) == 2 for x in d]):
                out_data.append(d)
                has_pairs = True
            elif all([isinstance(x, (int, str, float)) for x in d]):
                out_data.append(d)
                has_items = True
            else:
                raise ValueError("Invalid mixed-type series data.")

        if has_pairs and has_items:
            raise ValueError(
                "Use dicts/tuples or primitive values uniformly for all series."
            )
        return out_data


class ChartState(el.Element):
    is_collected = False
    x_axis = AxisProp("category")
    y_axis = AxisProp("value")
    rotate = BoolProp(False)
    tooltip = TooltipProp(None)
    color = ColorProp(None)
    data = ChartDataProp(None)
    series_names = TypedListProp(str, [])
    chart_type = OptionProp("bar", ("bar", "line", "scatter"))


class chart(combo):
    """
    See [Charts](/concepts/charts) for a charts tutorial.

    This component can render bar charts, line charts, and
    point/scatter charts.

    """

    def __init__(
        self,
        *data,  # Each separate series is a list/tuple
        chart_type="line",  # scatter, line, bar
        series_names=[],  # Used for tooltips and legends.
        x_axis="category",  # See parse_axis for valid format
        y_axis="value",
        rotate=False,
        tooltip=None,  # None, "axis", or "item" (where it's triggered)
        color=None,
        **props,  # width, height
    ):
        state = ChartState(
            x_axis=x_axis,
            y_axis=y_axis,
            rotate=rotate,
            data=data,
            series_names=series_names,
            tooltip=tooltip,
            chart_type=chart_type,
            color=color,
        )

        data = state.parsed_value("data")
        x_axis = state.parsed_value("x_axis")
        y_axis = state.parsed_value("y_axis")
        tooltip = state.parsed_value("tooltip")
        chart_type = state.parsed_value("chart_type")
        series_names = state.parsed_value("series_names")
        color = state.parsed_value("color")
        rotate = state.parsed_value("rotate")

        if (
            len(data) > 0
            and len(data[0]) > 0
            and type(data[0][0]) not in (list, tuple)
            and x_axis["type"] != "category"
        ):
            raise ValueError(
                "Invalid data. Use dict/tuple values or make axis type `category`."
            )

        series = parse_series(chart_type, data, series_names)

        if rotate:
            x_axis, y_axis = y_axis, x_axis

        theme = el.theme()

        chart = el.echart(
            options=dict(
                xAxis=x_axis,
                yAxis=y_axis,
                series=series,
                tooltip=tooltip,
                #
                # grid is used to remove spacing.
                grid=dict(top=8, right=8, bottom=0, left=0, containLabel=True),
                #
                # Use hyperdiv colors by default.
                color=color or dict(auto=True, seeds=[theme.primary_color]),
                axisTick=dict(show=False),
            ),
        )

        super().__init__(
            state,
            (
                chart,
                [
                    "data_clicked",
                    "width",
                    "height",
                    "min_width",
                    "min_height",
                    "grow",
                    "shrink",
                    "border",
                    "border_radius",
                    "padding",
                    "background_color",
                    "hover_background_color",
                    "active_background_color",
                    "font_color",
                    "font_weight",
                ]
                + el.input_props,
            ),
            init=props,
        )
