from .combo import combo
from .. import elements as el
from ..elements.props import PositiveIntProp


class TableState(el.state):
    page = PositiveIntProp(0)
    page_size = PositiveIntProp(0)


def data_table(data, length=None, id_field="id", page_size=20, **props):
    '''
    Renders the given data in a table, and renders a pagination
    component if the data size exceeds the page size.

    The data format accepted is a sequence of dictionaries with
    identical keys. The keys will be rendered as the table header.

    The data can be supplied either as an array containing the total
    data, or as a function. Here's an array example:

    ```py
    data = [
      dict(id=1, name="Steve", age=31),
      dict(id=2, name="Angie", age=22),
      dict(id=3, name="Rob", age=60)
    ]

    hd.data_table(data)
    ```

    Here's a function example:

    ```py
    data = [
      dict(id=1, name="Steve", age=31),
      dict(id=2, name="Angie", age=22),
      dict(id=3, name="Rob", age=60)
    ]

    def get_users(start_position, end_position):
        """Fetch users, e.g. from a database."""
        global data
        return data[start_position:end_position]

    def get_users_length():
        """Fetch the length of the users data set."""
        global data
        return len(data)

    hd.data_table(get_users, length=get_users_length())
    ```

    Note that when providing a function, you *must* provide a `length`
    value.

    By default, `data_table` assumes that there is a key in each
    dictionary, named `id`, that provides a unique identifier for each
    row in the data.

    If the ID key is not named `id`, you can provide its name with the
    `id_field` arg:

    ```py
    data = [
      dict(key=1, name="Steve", age=31),
      dict(key=2, name="Angie", age=22),
      dict(key=3, name="Rob", age=60)
    ]

    hd.data_table(data, id_field="key")
    ```

    Alternatively, the ID field can be a function that inputs a row
    and returns the row's ID:

    ```py
    data = [
      dict(key=1, name="Steve", age=31),
      dict(key=2, name="Angie", age=22),
      dict(key=3, name="Rob", age=60)
    ]

    def get_id(row):
        return row['key']

    hd.data_table(data, id_field=get_id)
    ```

    Finally, `hd.data_table` automatically renders a pagination
    component at the bottom of the table, if the `page_size` is
    smaller than the data length:

    ```py
    data = [
      dict(id=1, name="Steve", age=31),
      dict(id=2, name="Angie", age=22),
      dict(id=3, name="Rob", age=60),
      dict(id=4, name="Jim", age=31),
      dict(id=5, name="Lisa", age=22),
      dict(id=6, name="Bill", age=60)
    ]

    hd.data_table(data, page_size=2)
    ```
    '''

    if not callable(data):
        length = len(data)
        if not length:
            return
    else:
        if not isinstance(length, int) and length <= 0:
            raise Exception("length must be a positive int.")

    def get_data(start, end):
        if callable(data):
            return data(start, end)
        else:
            return data[start:end]

    def get_id_value(row):
        try:
            if callable(id_field):
                return id_field(row)
            else:
                return row[id_field]
        except Exception as e:
            raise ValueError(f"Could not retrieve ID from row: {str(e)}")

    state = TableState(page=0, page_size=min(page_size, 50))

    last_page = (length - 1) // state.page_size

    if state.page > last_page:
        state.page = last_page

    first_row = get_data(0, 1)[0]
    header = first_row.keys()

    start = state.page_size * state.page
    end = min(start + state.page_size, length)

    with el.box(**props) as wrapper:
        # Wrap the table in a container that scrolls horizontally, if
        # the table is wide.
        with el.box(horizontal_scroll=True, font_size="small"):
            with el.table():
                with el.table_head():
                    with el.table_row():
                        for h in header:
                            with el.scope(h):
                                el.table_cell(h, padding=8)
                with el.table_body():
                    for row in get_data(start, end):
                        with el.scope(get_id_value(row)):
                            with el.table_row():
                                for header, value in row.items():
                                    with el.scope(header):
                                        with el.table_cell(
                                            padding=8,
                                            height="fit-content",
                                            wrap_text=False,
                                            max_width=400,
                                            horizontal_scroll=True,
                                        ):
                                            el.plaintext(value)

        if length > state.page_size:
            # Render the paginator component
            with el.hbox(
                border=[0, 1, 1, 1],
                padding=8,
                justify="space-between",
                wrap="wrap",
            ):
                if el.text(
                    el.Icons.ChevronLeftSolid,
                    "Previous Page",
                    cursor="pointer",
                    font_color=(
                        el.highlight(el.theme().font_color, 200)
                        if state.page == 0
                        else None
                    ),
                ).clicked:
                    if state.page > 0:
                        state.page -= 1

                el.text(f"{start+1}-{end} / {length}")

                if el.text(
                    "Next Page",
                    el.Icons.ChevronRightSolid,
                    cursor="pointer",
                    font_color=(
                        el.highlight(el.theme().font_color, 200)
                        if state.page == last_page
                        else None
                    ),
                ).clicked:
                    if state.page < last_page:
                        state.page += 1

    return combo(wrapper, (state, ["page_size"]), init=props)
