from functools import cache
from .context import Context


def to_caml_case(name):
    parts = name.split("_")
    return "".join([parts[0], *[part.capitalize() for part in parts[1:]]])


class Prop:
    """
    The base class for all Prop classes. It is untyped and not
    recommended for defining props, unless they are truly untyped.
    """

    def __init__(self, value, is_event_prop=False):
        self.is_event_prop = is_event_prop
        self.value = value

    def parse(self, value):
        return value


class RemoveProp(Prop):
    def __init__(self):
        pass

    def parse(self, value):
        raise Exception("Invalid call `parse()` on RemoveProp")


class DynamicProp:
    def __init__(self, parent_element, name, static_prop):
        self.parent_element = parent_element
        self.name = name
        self.static_prop = static_prop

        self.value = static_prop.value
        self.is_event_prop = static_prop.is_event_prop

        self._cached_parse = None
        self.mutated = False

    @property
    def context(self):
        return self.parent_element._context

    def parse(self, value):
        # Parses the prop if the value has changed.
        # Otherwise it returns the cached parsed value.
        # Assumes that values can be compared using == and !=
        if callable(value):
            value = value(self.parent_element)
        if not self._cached_parse or value != self._cached_parse[0]:
            self._cached_parse = (value, self.static_prop.parse(value))

        return self._cached_parse[1]

    def init(self, value):
        if not self.mutated:
            self.parse(value)
            self.value = value

    def get_value(self):
        in_user_code = self.context.in_user_code()

        if in_user_code and self.is_event_prop:
            # If we're an event prop like `clicked`, etc., we record
            # user-code accesses into the read set.
            self.context.read_event_props.add(self)

        if callable(self.value):
            return self.value(self.parent_element)
        return self.value

    def _register_change(self, new_value):
        in_user_code = self.context.in_user_code()
        if in_user_code and new_value != self.value:
            self.context.frame_had_changes.set()

    def set_value(self, value):
        in_user_code = self.context.in_user_code()

        self.parse(value)

        if in_user_code and self.is_event_prop:
            # Event props are read-only from user code. Cannot write
            # to `clicked`, etc.
            raise Exception(f"Cannot write to even prop '{self.name}' from user code.")

        if self.is_event_prop:
            # We track mutated event props (clicked, etc.) so we can
            # reset them at the end of the frame.
            self.context.mutated_event_props.add(self)

        self._register_change(value)

        self.mutated = True
        self.value = value

    def reset(self):
        self._register_change(self.static_prop.value)
        self.value = self.static_prop.value
        self.mutated = False


def get_element_name(element_class):
    try:
        return element_class.name
    except AttributeError:
        # If a name is not set on the class, fall back to the class's
        # name.
        return element_class.__name__


class ElementCacheMetaclass(type):
    def __call__(cls, *args, **kwargs):
        context = Context.current()

        if cls.is_singleton:
            # TODO: A hash of the name would avoid possible collisions
            # with other user-supplied keys.
            key = get_element_name(cls)
        else:
            key = context.create_element_key(kwargs.pop("key", None))

        comp = context.element_state.get_element(key)
        if not comp:
            comp = super(ElementCacheMetaclass, cls).__call__(*args, key=key, **kwargs)
            context.element_state.set_element(comp)

        comp._init(*args, **kwargs)

        return comp


@cache
def collect_static_props(klass):
    """Walk up the class hierarchy, starting at klass, and collect
    all the `Prop`s into a flat dict.
    """
    props = dict()

    def collect_props(klass):
        """Post-order traversal makes sure that when we update the `props`
        dict, a prop in a subclass will override a prop with the
        same name in a superclass.
        """
        for b in klass.__bases__:
            if b != Element and issubclass(b, Element):
                collect_props(b)
        for k, v in vars(klass).items():
            if isinstance(v, RemoveProp):
                props.pop(k, None)
            elif isinstance(v, Prop):
                props[k] = (v, klass)

    collect_props(klass)
    return props


class PropAccessor:
    def __init__(self, props):
        self._props = props

    def __getattribute__(self, key):
        try:
            is_prop = key != "_props" and key in self._props
        except AttributeError:
            is_prop = False

        if is_prop:
            return self._props[key].get_value()
        else:
            return super().__getattribute__(key)

    def __setattr__(self, key, value):
        try:
            is_prop = key != "_props" and key in self._props
        except AttributeError:
            is_prop = False

        if is_prop:
            return self._props[key].set_value(value)
        else:
            return super().__setattr__(key, value)


class Element(PropAccessor, metaclass=ElementCacheMetaclass):
    # Defaults
    is_collected = True
    is_container = False
    is_singleton = False

    def __init__(self, *args, **kwargs):
        self._context = Context.current()
        self.key = kwargs.pop("key")

        self.name = get_element_name(self.__class__)
        if self.is_container:
            self.children = []

        self.was_collected = False

        def reset_collected():
            self.was_collected = False

        self._context.bus.on("frame_start", reset_collected)

        static_props = collect_static_props(self.__class__)

        dynamic_props = dict()
        by_id = dict()  # Account for aliasing
        for name, (static_prop, _) in static_props.items():
            prop_id = id(static_prop)
            if prop_id not in by_id:
                by_id[prop_id] = DynamicProp(self, name, static_prop)
            dynamic_props[name] = by_id[prop_id]

        super().__init__(dynamic_props)

    #
    # Methods that are designed to be overridden by element
    # writers. They all take as input the args that the user supplies
    # when invoking the element.
    #

    def args_to_props(self, *args, **kwargs):
        """Given the args/kwargs passed by the user to the element (minus
        the key kwarg), return a dict from prop names to prop values
        that will be used to initialize the props.

        By default, args is ignored and kwargs is assumed to map valid
        prop names to values.
        """
        return kwargs

    def on_render(*args, **kwargs):
        """Override this to do work after the element has been fully set
        up. This method is called on each frame in which the element
        is rendered.
        """
        pass

    #
    # The rest of these methods should in general not be overridden:
    #

    def _init_clear(self):
        if self.is_singleton and (not self.is_collected or self.was_collected):
            return
        self.clear()

    def _init_collect(self):
        if self.is_collected:
            if self.is_singleton and (not self.is_collected or self.was_collected):
                return
            self._context.collector_stack.collect(self)
            self.was_collected = True

    def _init_props(self, *args, **kwargs):
        init_values = self.args_to_props(*args, **kwargs)

        for prop_name, prop_value in init_values.items():
            prop = self._props.get(prop_name)
            if not prop:
                raise Exception(f'Unexpected property "{prop_name}" on "{self.name}".')
            prop.init(prop_value)

    def _init(self, *args, **kwargs):
        self._init_clear()
        self._init_collect()
        self._init_props(*args, **kwargs)

        self.on_render(*args, **kwargs)

    def mutated(self, propname=None):
        """[public] If `propname` is given, return `True` if that prop is
        mutated. If `propname` is not given, return `True` if *any*
        prop belonging to this element is mutated.
        """
        if propname:
            prop = self._props.get(propname)
            return prop and prop.mutated
        else:
            for prop in self._props.values():
                if prop.mutated:
                    return True
            return False

    def parsed_value(self, propname):
        prop = self._props.get(propname)
        if not prop:
            raise ValueError(f"Prop {propname} does not exist on {self.name}")
        return prop.parse(prop.value)

    def reset_prop(self, propname):
        """[public] Reset prop with the given name. Resetting a prop will set
        it to its original, default value, and clear its `mutated`
        state.
        """
        self._props[propname].reset()

    def clear(self):
        """[public] If this is a container element, remove all children
        elements from the container.  If not a container, calling this
        method has no effect.
        """
        if self.is_container:
            self.children.clear()

    def to_dom(self):
        dom = dict(
            _dom=True,
            name=self.name,
            key=self.key,
            **{
                to_caml_case(prop_name): prop.parse(prop.get_value())
                for prop_name, prop in self._props.items()
                # We don't send event props to the UI.
                if not prop.is_event_prop
            },
        )
        if self.is_container:
            dom["children"] = [element.to_dom() for element in self.children]
        return dom

    def __enter__(self):
        if not self.is_container:
            raise Exception(
                f"A '{self.name}' cannot be used in a `with` context "
                "because it's not a container."
            )
        self._context.collector_stack.push(self.children)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._context.collector_stack.pop()

    def __str__(self):
        props = " ".join(
            [f"{prop.name}={prop.get_value()}" for prop in self._props.values()]
        )
        return f"<{self.name} {props}>"

    def __repr__(self):
        return str(self)
