from .text_utils import markdown_with_code_examples


def application_state():
    markdown_with_code_examples(
        """
        # Application State

        Hyperdiv apps are composed of stateful UI components, and user
        interactions can cause changes in some of those props, causing
        the application to re-render.

        But often, we need application-specific custom state, separate
        from the UI components and their props. The
        [state](/components/state) component exists for this
        purpose. Here's a simple example where we build a click counter:

        ```py
        # Initialize state
        state = hd.state(count=0)

        if hd.button("Click Me").clicked:
            # Update state
            state.count += 1

        # Render state
        hd.text("You clicked", state.count, "times.")
        ```

        ## Untyped State

        `state` is extremely similar to a UI component like
        [checkbox](/components/checkbox). Like `checkbox`, `state` is
        also a collection of props. Props are initialized, accessed,
        and mutated, just like a normal UI component.

        The major difference is that `state`'s props by default are
        *untyped*, and are created *on the fly* when you initialize
        the component with keyword arguments. In `hd.state(count=0)`,
        the `count` prop is created on instantiation using the untyped
        [Prop](/proptypes#Prop) constructor.

        These are two different state objects with different sets of
        props:

        ```py
        state1 = hd.state(count=0)
        state2 = hd.state(name="Steve", age=25)
        ```

        And these props that are created on the fly are untyped:
        ```py
        state = hd.state(count=0)
        with hd.box(gap=8):
            if hd.button("Increment it").clicked:
               state.count += 1
            if hd.button("Modify it").clicked:
               state.count = "Steve"
            hd.text("Value:", state.count)
        ```

        Notice that we can set the count to a string value, and
        Hyperdiv does not complain. Attempting to increment it will
        then result in an error.

        ## Typed State

        Untyped state is very convenient for starting to whip up an
        application with minimal boilerplate. However, a more robust
        application will want to predefine *typed* props. This is done
        by making a class that inherits from `state`:
        ```py
        class MyState(hd.state):
            from hyperdiv.elements.props import IntProp

            count = IntProp(0)

        state = MyState()

        with hd.box(gap=8):
            if hd.button("Increment it").clicked:
               state.count += 1
            if hd.button("Modify it").clicked:
               state.count = "Steve"
            hd.text("Value:", state.count)        
        ```

        Notice that you can no longer succesfully click "modify
        it". The type `IntProp` prevents non-`int` values from being
        assigned into the prop.

        You can find the prop types that you can import from
        `hyperdiv.elements.props` [here](/proptypes).

        """
    )
