from .text_utils import markdown_with_code_examples


def box_layout():
    markdown_with_code_examples(
        """
        # Box Layout

        Hyperdiv supports powerful layout using arbitarily nested
        [boxes](/components/box). This documentation app's UI is built
        largely using boxes. The top-bar, the menu on the left and its
        items, and this content area are boxes.

        See [box](/components/box) for all the props you can use and
        their values.

        ## Box Layout Basics

        When laying content in boxes, some of the most important
        concepts to understand are direction, padding, border, gap,
        and dimensions.

        * `direction`: The direction of a box determines how children
        are laid out inside it. A [box](/components/box) lays out its
        children *vertically* and a [hbox](/components/hbox) lays out
        its children *horizontally*.

        * `gap`: The gap prop controls how much space to put between the
        children of a box.

        * `padding`: The padding prop controls the amount of space
        between the content (children) of the box and the sides/edges
        of the box.

        * `border`: The border prop lets you draw a visual border
        around the box.

        * `width` and `height`: Control the dimensions of a box.

        Here's an app that demonstrates the behaviors of these props
        in a vertical box, as well as a horizontal one:

        ```py
        with hd.box(gap=16):
            with hd.box():
                padding = hd.checkbox('Toggle Padding')
                border = hd.checkbox('Toggle Border')
                gap = hd.checkbox('Toggle Gap')
                width = hd.checkbox('Toggle Width')

            padding = 0 if padding.checked else 16
            border = 0 if border.checked else 1
            gap = 0 if gap.checked else 16
            width = "50%" if width.checked else None

            with hd.box(
                padding=padding,
                border=border,
                gap=gap,
                width=width,
            ):
                hd.button('Child 1')
                hd.button('Child 2')

            with hd.hbox(
                padding=padding,
                border=border,
                gap=gap,
                width=width,
            ):
                hd.button('Child 1')
                hd.button('Child 2')

        ```

        ## Box Styling

        Boxes can styled with background colors, font colors and
        sizes, and a variety of options for borders and padding.

        ```py
        with hd.box(
            width="50%",
            border=(10, "solid", "green"),
            border_radius=10,
            padding=40,
            background_color="white",
            hover_background_color="yellow",
            font_color="red",
            font_size="x-large"
        ):
            hd.text("Hello")
        ```

        We should note that the props that control border, border
        radius, and padding can be used for independent control of the
        four corners and the four sides of a box:

        ```py
        with hd.box(
            width="50%",
            border=[0, 0, (10, "solid", "green"), 0],
            border_radius=[0, 0, 10, 10],
            padding=[10, 40, 10, 40],
            background_color="white",
            hover_background_color="yellow",
            font_color="red",
            font_size="x-large"
        ):
            hd.text("Hello")
        ```

        ## More Advanced Box Layout

        Boxes can be nested in powerful ways to achieve complex
        layouts. Here's a sketch for how to lay out an app with a menu
        on the left.

        ```py
        def top_bar():
            with hd.box(
                border=[0, 0, 1],
                padding=16,
            ):
                hd.text('My Application')

        def menu(items):
            def menu_item(item):
                with hd.box(
                    hover_background_color=hd.theme().surface_color,
                    cursor="pointer",
                    padding=8
                ):
                    hd.text(item)

            with hd.box(
                border=[0, 1],
                width=200,
                padding=16,
            ):
                for item in items:
                    with hd.scope(item):
                        menu_item(item)

        def content():
            with hd.box(padding=16, gap=16):
                hd.text("# Hello")
                hd.text("Welcome to my app.")

        with hd.box(border=1):
            top_bar()
            with hd.hbox():
                menu(["Introduction", "Overview", "Getting Started"])
                content()
        ```

        Here's the same app, with exagerrated borders and gaps so you
        can visually see the nesting structure: 

        ```py
        def top_bar():
            with hd.box(
                border=1,
                padding=16,
            ):
                hd.text('My Application')

        def menu(items):
            def menu_item(item):
                with hd.box(
                    hover_background_color=hd.theme().surface_color,
                    cursor="pointer",
                    padding=8,
                    border=1
                ):
                    hd.text(item)

            with hd.box(
                border=1,
                width=200,
                padding=16,
                gap=8
            ):
                for item in items:
                    with hd.scope(item):
                        menu_item(item)

        def content():
            with hd.box(padding=16, gap=16, border=1):
                hd.text("# Hello")
                hd.text("Welcome to my app.")

        with hd.box(border=1, padding=8, gap=8):
            top_bar()
            with hd.hbox(border=1, padding=8, gap=8):
                menu(["Introduction", "Overview", "Getting Started"])
                content()
        ```

        """
    )
