import hyperdiv as hd
from .text_utils import markdown_with_code_examples


def component_link(name):
    return f"[`{name}`](/components/{name})"


def charts():
    markdown_with_code_examples(
        f"""
        # Charts

        Hyperdiv supports powerful chart components via the [Apache
        ECharts](https://echarts.apache.org/en/index.html) library.

        Hyperdiv exposes high level chart components
        {component_link('bar_chart')}, {component_link('line_chart')},
        and {component_link('scatter_chart')}. Their interfaces are
        identical, and in fact they are all wrappers around
        {component_link('chart')}. And {component_link('chart')}
        itself is a high-level interface on top of the uber-powerful
        low-level component {component_link('echart')}.

        ## Basic Examples

        In this section, we start with a basic chart and progressively
        add features. You can use the following dropdown to control
        which chart type is used in the examples in this section:

        """
    )

    with hd.box(width=200):
        chart_type = hd.select("Chart Type", "bar", ("bar", "line", "scatter")).selected

    markdown_with_code_examples(
        f"""
        The simplest way to get a chart:

        ```py
        hd.{chart_type}_chart([6, 2, 3, 4])
        ```

        We can add more series by passing additional lists of values:

        ```py
        hd.{chart_type}_chart(
            [6, 2, 3, 4],
            [6, 8, 3, 10],
            [10, 9, 16, 22]
        )
        ```

        Notice that the values for the x-axis are integers, chosen
        automatically in increasing order, by default. But we can
        customize the x-axis:

        ```py
        hd.{chart_type}_chart(
            [6, 2, 3, 4],
            [6, 8, 3, 10],
            [10, 9, 16, 22],
            x_axis=["Oats", "Corn", "Wheat", "Rice"]
        )
        ```

        Notice that the order of the x-axis labels matches the order
        of the data points in each series.

        We can then enable tooltips, so you can see detail data when you
        hover the mouse over the chart:

        ```py
        hd.{chart_type}_chart(
            [6, 2, 3, 4],
            [6, 8, 3, 10],
            [10, 9, 16, 22],
            x_axis=["Oats", "Corn", "Wheat", "Rice"],
            tooltip="axis"
        )
        ```

        The other possible value for tooltip is `"item"`, which would give
        you info about each individual point, instead of grouping all the
        points for each axis item.

        Notice that the series names, as rendered in the tooltips, are
        also just a sequence of integers by default. We can customize the
        series names:

        ```py
        hd.{chart_type}_chart(
            [6, 2, 3, 4],
            [6, 8, 3, 10],
            [10, 9, 16, 22],
            x_axis=["Oats", "Corn", "Wheat", "Rice"],
            series_names=["Harvested", "Remaining", "2023 Forecast"],
            tooltip="axis"
        )
        ```

        Finally, we can perform actions when users click the chart's
        points. In the example below, we just store the data of the
        point that was last clicked, so we can display it.

        ```py
        chart = hd.{chart_type}_chart(
            [6, 2, 3, 4],
            [6, 8, 3, 10],
            [10, 9, 16, 22],
            x_axis=["Oats", "Corn", "Wheat", "Rice"],
            series_names=["Harvested", "Remaining", "2023 Forecast"],
            tooltip="axis"
        )
        state = hd.state(last_clicked=None)
        if chart.data_clicked:
            state.last_clicked = chart.data_clicked
        hd.text("Last clicked", state.last_clicked)
        ```

        The `data_clicked` prop contains data for a single frame, and is
        automatically reset to `None` until there is another click. It
        has the following fields:

        * `series`: The series name as set by the `series_names` prop.
        * `data_index`: The position of the data series in the list of
          given data series, starting at `0`.
        * `data`: The value of the data point.
        * `name`: The name of the `x_axis` value corresponding to this
          data point.

        ## Other ways to specify data

        Instead of lists of y-values, we can specify data using dicts
        or tuples. When using dicts or tuples, the keys in the
        dictionary (or the first element in each tuple) represents
        the axis item to which the value is mapped.

        These are three equivalent ways to specify the same
        chart:

        ```py
        hd.bar_chart(
            dict(Apples=10, Oranges=20, Pears=11),
            dict(Apples=21, Oranges=50, Pears=30),
        )
        ```
        ```py
        hd.bar_chart(
            [("Apples", 10), ("Oranges", 20), ("Pears", 11)],
            [("Apples", 21), ("Oranges", 50), ("Pears", 30)]
        )
        ```
        ```py
        hd.bar_chart(
            [10, 20, 11],
            [21, 50, 30],
            x_axis=("Apples", "Oranges", "Pears")
        )
        ```

        ## Axis Types

        ### category

        By default, the `x_axis` type is `"category"`, which roughly
        represents a finite collection of possible items. When you
        specify `x_axis=("One", "Two")`, that specifies a 2-item
        category. If you don't specify an `x_axis` type, it defaults
        to `"category"`. If you don't use tuples or dicts to map data
        items to axis items, the data will be mapped to an
        automatically-generated axis item, starting with `0` and
        incrementing by `1`.

        ### value

        Another important type of axis is `"value"`. This type assumes
        the axis items are numbers laid out on the number line. Here's
        the difference between `"value"` and `"category"`:

        ```py
        hd.scatter_chart(
            [(0, 1), (3, 5), (2.5, 2)],
            x_axis="value"
        )

        hd.scatter_chart(
            [(0, 1), (3, 5), (2.5, 2)],
            x_axis="category"
        )
        ```

        Notice that in the `"value"` case, the points are sorted and
        laid out on an intelligently-generated x-axis that represents
        the number line. In the `"category"` case, the x-axis is taken
        verbatim from the data; the x-values are not sorted or
        interpreted as numbers.

        ### time

        The axis type `"time"` behaves very similarly to `"value"`,
        but instead of interpreting x-values as numbers, it interprets
        them as time values. Similarly to `"value"`, it intelligently
        generates the x-axis labels.

        ```py
        import time
        import datetime

        points = [
            (str(datetime.datetime(2014, 12, 8, 4, 38, 40)), 10),
            (str(datetime.datetime(2016, 8, 24, 1, 0, 21)), 21),
            (str(datetime.datetime(2011, 6, 8, 4, 12, 40)), 16),
            (str(datetime.datetime(2020, 4, 1, 4, 51, 21)), 11)
        ]
        hd.scatter_chart(points, x_axis="time")
        ```

        In addition to `datetime` strings, correct time values are
        also numbers of milliseconds since
        [Epoch](https://en.wikipedia.org/wiki/Unix_time). For example:

        ```py
        hd.line_chart([
            (1658466268671, 21),
            (1650176266166, 100),
            (1648556268671, 10),
        ], x_axis="time")
        ```

        ## The `echart` Component

        The chart types discussed above are implemented on top of the
        {component_link('echart')} component. This component exposes a
        raw interface to ECharts, so we don't recommend using it for
        normal use cases, but it also exposes the full power and
        customizability of ECharts.

        The component's `options` prop takes a nested dictionary that
        supports the options documented
        [here](https://echarts.apache.org/en/option.html). Here's a
        complex example copied from
        [here](https://echarts.apache.org/handbook/en/concepts/axis/).

        ```py
        options = dict(
          tooltip=dict(
            trigger='axis',
            axisPointer=dict(type='cross')
          ),
          legend=dict(),
          xAxis=[
            dict(
              type='category',
              axisTick=dict(
                alignWithLabel=True
              ),
              axisLabel=dict(
                rotate=30
              ),
              data=[
                'January',
                'February',
                'March',
                'April',
                'May',
                'June',
                'July',
                'August',
                'September',
                'October',
                'November',
                'December'
              ]
            )
          ],
          yAxis=[
            dict(
              type='value',
              name='Precipitation',
              min=0,
              max=250,
              position='right',
              axisLabel=dict(
                formatter='{{value}} ml'
              )
            ),
            dict(
              type='value',
              name='Temperature',
              min=0,
              max=25,
              position='left',
              axisLabel=dict(
                formatter='{{value}} °C'
              )
            )
          ],
          series=[
            dict(
              name='Precipitation',
              type='bar',
              yAxisIndex=0,
              data=[6, 32, 70, 86, 68.7, 100.7, 125.6, 112.2, 78.7, 48.8, 36.0, 19.3]
            ),
            dict(
              name='Temperature',
              type='line',
              smooth=True,
              yAxisIndex=1,
              data=[
                6.0,
                10.2,
                10.3,
                11.5,
                10.3,
                13.2,
                14.3,
                16.4,
                18.0,
                16.5,
                12.0,
                5.2
              ]
            )
          ]
        )
        hd.echart(options=options, height=350)
        """
    )
