import hyperdiv as hd
from .text_utils import code_example, markdown_with_code_examples


def notice(label=None, accent_color="yellow", icon=hd.Icons.TriangleExclamationSolid):
    container = hd.hbox(
        border=accent_color,
        padding=16,
        gap=16,
    )
    if label:
        with container:
            hd.icon(icon, font_color=accent_color)
            hd.text(label)
    else:
        return container


def getting_started():
    markdown_with_code_examples(
        """
        # Getting Started

        In this section we'll show the Hello World app and reveal some
        core concepts.

        ## Hello World

        Fundamentally, a Hyperdiv app is a function that lays out
        UI components on the screen. Here's how to set up the
        simplest possible app:
        """
    )

    hd.code(
        """
        # Import Hyeperdiv:
        import hyperdiv as hd

        # Define your app:
        def my_app():
            hd.text("Hello, World!")

        # Run the app:
        hd.run(my_app)
        """
    )

    markdown_with_code_examples(
        """
        You can paste this code into a file called, say,
        `test.py`, and then run it the usual way with `python
        test.py`:

        ```
        $ python test.py
        Open http://localhost:8888/
        ```
        The app will print a local url, which you can paste into the
        browser. To close the app, hit `Ctrl-C` at the terminal and
        close the browser tab.

        As you edit the app source code, Hyperdiv should automatically
        restart the app to show the new changes.

        """
    )

    notice(
        """ Throughout the rest of this documentation, we will show Hyperdiv
        code snippets with their rendered UI side-by-side. We assume
        the shown code runs inside the body of the app function, and
        will no longer show the `import hyperdiv as hd` and `hd.run()`
        boilerplate that you see above.  We will refer to the
        `hyperdiv` module as `hd` throughout.
        """
    )

    markdown_with_code_examples(
        """
        ## Components and Props

        In the example above, `hd.text("Hello, World!")` shows
        the Hyperdiv [text](/components/text) component in action.

        A Hyperdiv component functions like a normal Python object
        with attributes. But in Hyperdiv, a component's attributes are
        smarter than normal attributes, and throughout this
        documentation we call them **props**.

        Any prop of any component can be initialized by passing it to
        the component constructor as a "keyword argument", and also
        accessed and mutated as an attribute:

        ```py
        # Initialize it:
        ch = hd.checkbox("Check Me", checked=True)
        # Access it:
        hd.text(ch.checked)
        # Mutate it:
        if hd.button("Flip it").clicked:
            ch.checked = not ch.checked
        ```

        Notice that some components, like `text` and `checkbox` (and
        many others) also take prop values as normal arguments,
        instead of keyword arguments. Those components perform an
        internal translation where they take the non-keyword arguments
        and use them to initialize props. In the case of `text` and
        `checkbox`, they initialize their respective `label` prop
        using the non-keyword argument.

        ```py
        hd.checkbox("Check Me")
        # Is equivalent to
        hd.checkbox(label="Check Me")
        ```

        ## Interactivity

        When any prop of any component in the app changes, Hyperdiv
        **re-runs the app function**. When the app re-runs, the changes in
        props can cause it to render differently.

        Here's an example:
        ```py
        if hd.checkbox("Show greeting").checked:
            hd.text("Hello, World!")
        ```

        The value of the `checked` prop of the checkbox starts off as
        `False`. When we click the checkbox in the UI, the value of
        the `checked` prop is set to `True`, causing the app to
        re-run.

        When it re-runs, it renders a different UI based on the value
        of the `checked` prop.

        ## Gotcha: Infinite Refreshes

        This code will cause Hyperdiv to re-run your app indefinitely.
        """
    )
    hd.code(
        """
        ch = hd.checkbox("Hello")
        ch.checked = not ch.checked
        """
    )
    markdown_with_code_examples(
        """
        This is because every time it re-runs, the `checked` prop is set
        to a different value, causing it to re-run again. Hyperdiv
        will detect this infinite loop and raise an error.

        Prop mutations should only happen in response to events, such
        as `clicked`. See [User Input](/concepts/user-input) for more.

        ## Gotcha: Loops

        Due to the way Hyperdiv works internally, a loop like the
        following will raise an error:

        ```py
        my_data = ["One", "Two", "Three"]
        for item in my_data:
            hd.text(item)
        ```

        When looping over data, the body of a loop needs to be wrapped
        in a [scope](/components/scope) with a unique ID for every
        loop iteration. In this case, since we have no repeating data
        items, we can use the item as the scope ID:

        ```py
        my_data = ["One", "Two", "Three"]
        for item in my_data:
            with hd.scope(item): # Use the item as the scope ID
                hd.text(item)
        ```
        """
    )
