import hyperdiv as hd
from .text_utils import markdown_with_code_examples


def loops():
    markdown_with_code_examples(
        """
        # Looping over Data

        When iterating over data and rendering each data item, we need
        to be careful to **scope** the loop bodies. Otherwise Hyperdiv
        will raise a `Duplicate Key` error.

        ```py
        fruit = ["cherry", "pear", "grapefruit"]
        for item in fruit:
            hd.text("I ate a", item)
        ```

        To fix this, we need to use `scope` to scope the body of the
        loop to a unique ID. In the example above, we can assume the
        fruit names are unique, and use the fruit name as the scope
        ID:

        ```py
        fruit = ["cherry", "pear", "grapefruit"]
        for item in fruit:
            with hd.scope(item):
                hd.text("I ate a", item)
        ```

        ## Why?

        The reason for the `Duplicate Key` error is that Hyperdiv
        automatically generates a unique ID for each hyperdiv
        component in the UI, by using the *line number* where the
        component is instantiated in the Python file.

        The reason Hyperdiv needs these unique IDs is complex, but
        simply put it's so it can look up the correct component on
        each re-render of the application.

        Let's look at the example above, again:
        """
    )

    hd.code(
        """
        fruit = ["cherry", "pear", "grapefruit"]
        for item in fruit:
            hd.text("I ate a", item)
        """,
        show_line_numbers=True,
    )

    markdown_with_code_examples(
        """
        When first executing the code on line 3, Hyperdiv generates a
        unique ID for the `hd.text` element instantiated there, based
        on the line number alone.  On the next iteration, when
        Hyperdiv encounters `hd.text` again, it attempts to generate a
        unique ID for that line number, but fails because a unique ID
        for that line number was already generated.

        By wrapping the loop body in `hd.scope`, we give Hyperdiv
        extra information the extra uniqueness information it needs
        for generating its unique ID, in addition to just the line
        number.

        The rule of thumb is: When looping over data, always wrap the
        loop body in `hd.scope`, with a unique ID for each iteration.

        ## Scope IDs

        In the example above, coming up with a unique ID for each
        scope is easy enough: use the data items themselves, since
        they're unique.

        In other situations, the data items may not be unique. In that
        case, using the data items as scope IDs does not work:
        
        ```py
        fruit = ["cherry", "cherry", "pear"]
        for item in fruit:
            with hd.scope(item):
                hd.text("I ate a", item)
        ```

        As a quick workaround, we can use the index of each item in
        the data array:

        ```py
        fruit = ["cherry", "cherry", "pear"]
        for i, item in enumerate(fruit):
            with hd.scope(i):
                hd.text("I ate a", item)
        ```

        This approach works very well for static data, but will break
        down when:

        * the data items are stateful
        * the list changes over time

        Here's a contrived example:

        ```py
        state = hd.state(fruit=["cherry", "cherry", "pear"])

        def delete_at(idx):
            global state
            state.fruit = [
                fruit
                for i, fruit in enumerate(state.fruit)
                if i != idx
            ]

        with hd.box(gap=8):
            for i, item in enumerate(state.fruit):
                with hd.scope(i):
                    with hd.box(border=1, gap=8, padding=8):
                        hd.text(item)
                        hd.checkbox("Select it")
                        if hd.button("Delete").clicked:
                            delete_at(i)
        ```

        In the example above, click a checkbox to select the second
        "cherry" item, and then delete that same item. Notice that the
        checkbox stays checked, effectively implicitly selecting the
        next "pear" item, even though we never selected that item.

        This is because we scoped the loop body to the item position
        in the array, and the relationship between the item and its
        position changes.

        ## Major Takeaways:

        * Always wrap loop bodies in `hd.scope`, with a unique ID for
          each item.

        * You may use the item's position in the array as the unique
          ID, if the array never changes.

        * Otherwise, you need a reliably unique ID for each data item.

        When working with databases, this is straightforward enough:
        just use each item's primary key as the scope ID.

        """
    )
