import hyperdiv as hd

from hyperdiv.components.uilist import uilist_formatter

from .text_utils import markdown_with_code_examples
from .prop_tables import prop_tables
from .method_tables import method_tables

from .user_input import user_input
from .getting_started import getting_started
from .overview import overview
from .markdown_page import markdown
from .charts import charts
from .box_layout import box_layout
from .application_state import application_state
from .tables import tables
from .modals import modals
from .routing import routing
from .modular_apps import modular_apps
from .loops import loops
from .json_page import json_page


def proptypes():
    # TODO: add the props that are not in props.py
    hd.text("# Prop Types")
    proptypes = {
        propname: proptype
        for propname, proptype in hd.elements.props.__dict__.items()
        if propname.endswith("Prop")
    }
    from hyperdiv.components.charts import AxisProp

    # TODO: Add all props that are not in props.py
    proptypes["AxisProp"] = AxisProp

    for propname, proptype in proptypes.items():
        with hd.scope(proptype):
            with hd.box(gap=4):
                hd.anchor(propname)
                hd.text(f"#### {propname}")
                hd.text(proptype.__doc__)


def top_bar():
    """The top navigation bar. Currently nothing up here but the app
    title.
    """
    with hd.hbox(
        width="100%",
        padding=16,
        gap=16,
        align="center",
        height=48,
        border=(None, None, hd.theme().input_border, None),
    ):
        hd.text("**Hyperdiv** Docs")


def component_page(component_name, *args, **kwargs):
    def fn():
        klass = eval(f"hd.{component_name}")
        with hd.box():
            hd.text(f"# {component_name}")
            if klass.__doc__:
                markdown_with_code_examples(klass.__doc__)

        with hd.box(gap=32):
            prop_tables(component_name, *args, **kwargs)
            method_tables(component_name)

    return fn


def spec(name, *args, **kwargs):
    """A component name along with the *args and **kwargs used to
    instantiate it. The args/kwargs are only relevant for
    non-elements: components that are combos or functions.

    TODO: All components should probably be instantiatable without
    arguments.

    """
    return (name, args, kwargs)


components_list = [
    spec("bar_chart"),
    spec("chart"),
    spec("box"),
    spec("button"),
    spec("checkbox"),
    spec("clipboard_button", ""),
    spec("code"),
    spec("dark_mode_toggle"),
    spec("date"),
    spec("dialog", ""),
    spec("drawer"),
    spec("drawer_element"),
    spec("echart"),
    spec("form"),
    spec("head"),  # need to document set_title, fonts, etc
    spec("icon"),
    spec("icon_button", hd.Icons.BaconSolid),
    spec("image"),
    spec("json"),
    spec("label_container"),
    spec("line_chart"),
    spec("location"),
    spec("modal"),
    spec("password"),
    spec("plaintext"),
    spec("radio"),
    spec("radio_group", "", "", [""]),
    spec("scatter_chart"),
    spec("search"),
    spec("select", "", "", [""]),
    spec("simple_dialog"),
    spec("slot"),
    spec("table"),
    spec("table_head"),
    spec("table_body"),
    spec("table_foot"),
    spec("table_row"),
    spec("table_cell"),
    spec("tabs", ()),
    spec("text"),
    spec("text_input"),
    spec("theme"),
    spec("data_table", [dict(id=1)]),
    spec("tray"),
    spec("tray_element"),
    spec("uilist", [], uilist_formatter),
    # TODO: should these base classes be exposed?
    spec("Element"),
    spec("FormInput"),
    spec("InputElement"),
    spec("LabelElement"),
    spec("StyledElement"),
    # TODO: should these element classes (which are
    # shadowed by combos of the same name) be exposed?
    spec("elements.checkbox"),
    spec("elements.clipboard_button"),
    spec("elements.date"),
    spec("elements.modal"),
    spec("elements.password"),
    spec("elements.search"),
    spec("elements.select"),
    spec("elements.text_input"),
    # combo
    # scope
    # state
]


def components():
    hd.text("# Components")
    hd.text(
        "\n\n".join([f"[{name}](/components/{name})" for name, _, _ in components_list])
    )


menu_items = [
    dict(
        title="Intro",
        children=[
            dict(
                title="Overview",
                route="/",
                fn=overview,
            ),
            dict(
                title="Getting Started",
                route="/getting-started",
                fn=getting_started,
            ),
        ],
    ),
    dict(
        title="Concepts",
        children=[
            dict(
                title="Text & Markdown",
                route="/concepts/markdown",
                fn=markdown,
            ),
            dict(
                title="Box Layout",
                route="/concepts/box-layout",
                fn=box_layout,
            ),
            dict(
                title="Application State",
                route="/concepts/application-state",
                fn=application_state,
            ),
            dict(
                title="User Input and Forms",
                route="/concepts/user-input",
                fn=user_input,
            ),
            dict(
                title="Charts",
                route="/concepts/charts",
                fn=charts,
            ),
            dict(
                title="Tables",
                route="/concepts/tables",
                fn=tables,
            ),
            dict(
                title="Exploring Data",
                route="/concepts/json",
                fn=json_page,
            ),
            dict(
                title="Modal Dialogs",
                route="/concepts/dialogs",
                fn=modals,
            ),
            dict(
                title="Routing",
                route="/concepts/routing",
                fn=routing,
            ),
            dict(
                title="Building Modular Apps",
                route="/concepts/building-modular-apps",
                fn=modular_apps,
            ),
            dict(
                title="Looping Over Data",
                route="/concepts/looping-over-data",
                fn=loops,
            ),
        ],
    ),
    dict(
        title="Components",
        children=[
            dict(
                title="All Components",
                route="/components",
                fn=components,
            ),
            dict(
                title="All Props",
                route="/proptypes",
                fn=proptypes,
            ),
        ],
    ),
]


def tree_menu(items):
    for item in items:
        title = item["title"]
        with hd.scope(title):
            if "children" in item and len(item["children"]) > 0:
                hd.text(f"### {title}")
                with hd.box(padding=[4, 0, 16, 0]):
                    tree_menu(item["children"])

            else:
                location = hd.location()
                route = item["route"]

                if route == hd.location().path:
                    bgcolor = hd.highlight(hd.theme().background_color, 25)
                else:
                    bgcolor = None

                if "hide" not in item or not item["hide"]:
                    item = hd.text(
                        title,
                        background_color=bgcolor,
                        padding=8,
                        cursor="pointer",
                        font_size="small",
                    )
                    if item.clicked:
                        location.go(route)


def side_menu():
    """The lefthand menu with the list of components.
    When elements are selected the router updates
    and corresponding content is rendered."""
    with hd.box(
        height="100%",
        shrink=0,
        width=200,
        background_color=hd.theme().surface_color,
        padding=8,
        vertical_scroll=True,
    ):
        tree_menu(menu_items)


def get_menu_page(menu, route):
    for i in menu:
        if "children" in i:
            page = get_menu_page(i["children"], route)
            if page:
                return page
        elif i["route"] == route:
            return i["fn"]


components_routes = {
    f"/components/{name}": component_page(name, *args, **kwargs)
    for name, args, kwargs in components_list
}


def get_non_menu_page(route):
    if route in components_routes:
        return components_routes[route]


def main():
    location = hd.location()

    hd.set_title("Hyperdiv Documentation")
    top_bar()

    # TODO: I'm not sure why I needed to set this hbox's vertical_scroll to True
    # But otherwise the entire page would scroll.
    with hd.hbox(vertical_scroll=True, height="100%"):
        side_menu()
        # Use the scope to ensure each page has it's own
        # Scrollable container.
        with hd.scope(location.path):
            with hd.box(
                grow=1,
                padding=[16, 40, 400, 40],
                vertical_scroll=True,
            ):
                with hd.box(
                    gap=32,
                    min_width=600,
                    max_width=1000,
                ):
                    fn = get_menu_page(menu_items, location.path)
                    if not fn:
                        fn = get_non_menu_page(location.path)
                    if fn:
                        fn()
                    else:
                        hd.text("# 404 Not Found.")
                        hd.text(f"The path {location.path} doesn't point to anything.")
