from .text_utils import markdown_with_code_examples


def modular_apps():
    markdown_with_code_examples(
        """
        # Building Modular Apps

        In frontend frameworks like React, Vue, and Svelte, you'll
        typically break an application up into smaller "components"
        and then compose those into complex user interfaces. The
        "components" in those frameworks are special units of
        composition, different than the normal functions and classes
        we're used to. They're defined differently, and they use
        special syntax.

        In stark contrast to those frameworks, Hyperdiv doesn't have
        any special notion of "component". Normal Python functions and
        classes are used for breaking an app into modular units.

        In those frameworks you use an additional, limited HTML/JSX
        syntax to render components. For example, to render a `Person`
        component in React you'd do:

        ```markup
        <div>
          <Person name="Steve" age=20 />
        </div>
        ```

        In Hyperdiv, you just call the function, or instantiate the
        class, of your component, as you normally would in a normal
        Python program.

        Here's how we define and call a reusable component in
        Hyperdiv:

        ```py
        def person(name, age):
            with hd.box(border=1, padding=16):
                hd.text("**Name:**", name)
                hd.text("**Age:**", age)

        # Now let's render some people:
        with hd.box(gap=8):
            person("Steve", 34)
            person("Mary", 26)
            person("Joe", 31)
        ```

        ## Internal State

        Components may also use internal state. Here's an example
        where each instance of `person` renders a `Delete` button,
        that once clicked, will cause the component to be grayed out
        forever.

        ```py
        def person(name, age):
            state = hd.state(deleted=False)
            with hd.hbox(
                border=1,
                padding=16,
                justify="space-between",
                background_color=(
                    hd.theme().surface_color 
                    if state.deleted else None
                )
            ):
                with hd.box():
                    hd.text("**Name:**", name)
                    hd.text("**Age:**", age)
                if not state.deleted:
                    if hd.button("Delete").clicked:
                        state.deleted = True

        # Now let's render some people:
        with hd.box(gap=8):
            person("Steve", 34)
            person("Mary", 26)
            person("Joe", 31)
        ```

        ## Exposing State
        
        Components that are meant to be reused may expose their
        internal state. For example, the component above may not
        render its own `Delete` button, but instead expose a `delete`
        method that the user of the component may call. The user of
        the component may then render their own UI for deleting a
        user.

        ```py
        class Person:
            def __init__(self, name, age):
                self.name = name
                self.age = age
                self.state = hd.state(deleted=False)

                with hd.hbox(
                    border=1,
                    padding=16,
                    background_color=(
                        hd.theme().surface_color 
                        if self.state.deleted else None
                    )
                ):
                    with hd.box():
                        hd.text("**Name:**", name)
                        hd.text("**Age:**", age)

            def delete(self):
                self.state.deleted = True


        with hd.box(gap=16):
            with hd.box(gap=8):
                users = [
                    Person("Steve", 34),
                    Person("Mary", 26),
                    Person("Joe", 31)
                ]

            user_to_delete = hd.select(
                "Choose the user to delete:",
                users[0].name,
                [u.name for u in users]
            ).selected

            if hd.button("Delete", user_to_delete).clicked:
                for u in users:
                    if u.name == user_to_delete:
                        u.delete()
        ```

        Notice that `name` and `age` were not set up as state props,
        but instead they are members of the class. This is because
        they are not mutable, or intended to be mutated in our app. If
        we were to mutate them in the app, they should be state props,
        so that why they are mutated, the app will re-render.

        Tangent: Also notice that we collect the components into a
        `users` list while simultaneously rendering the users. When
        calling `Person("Steve", 34)`, that object is rendered into
        the document, and then the person object is returned, which we
        collect into the list for later use.

        ## Exposing Content Containers

        In addition to exposing internal state, components can expose
        containers that users of the component can plug their own
        arbitrary content into.

        See [slot](/components/slot) for an example.

        """
    )
