import inspect
import hyperdiv as hd
from hyperdiv.core.element import collect_static_props, Element
from hyperdiv.core.context import paused_collector
from .table_utils import render_value, cell_style, header_row, qualname


def organize_props_by_class(element_or_combo, element_name, *args, **kwargs):
    by_class = dict()

    # If it's a standard element, collect the static props
    if inspect.isclass(element_or_combo) and issubclass(element_or_combo, Element):
        props = collect_static_props(element_or_combo)

    # Otherwise create an instance and get the props dynamically
    else:
        props = dict()
        with paused_collector():
            combo_instance = element_or_combo(*args, **kwargs)

        if not hasattr(combo_instance, "_props"):
            return dict()

        unorganized_props = combo_instance._props
        for propname, prop in unorganized_props.items():
            if prop.parent_element.__class__ == hd.state:
                props[propname] = (prop.static_prop, element_or_combo)
            else:
                static_props = collect_static_props(prop.parent_element.__class__)
                props[propname] = static_props[propname]

    # Store props by their source class
    for propname, (prop, propclass) in props.items():
        if propclass not in by_class:
            by_class[propclass] = []

        by_class[propclass].append((propname, prop))

    sorted_by_class = []
    for propclass, props in by_class.items():
        props = sorted(props, key=lambda p: p[0])
        sorted_by_class.append((propclass, props))

    # Sort the props in alphabetical order by class name,
    # But put the props defined directly by the element first.
    sorted_by_class = sorted(
        sorted_by_class,
        key=lambda p: ""
        if qualname(p[0]) == qualname(element_or_combo)
        else qualname(p[0]),
    )

    return sorted_by_class


def prop_table(props):
    from hyperdiv.elements.props import OptionProp

    for (propname, prop) in props:
        with hd.scope(propname):
            with hd.table_row():
                hd.table_cell(propname, **cell_style)
                proptype = type(prop).__name__
                with hd.table_cell(**cell_style, gap=8):
                    hd.text(f"[{proptype}](/proptypes#{proptype})")
                    if isinstance(prop, OptionProp):
                        options_text = "\n".join(
                            [f"* {render_value(o)}" for o in prop.options]
                        )
                        hd.text(f"Options:\n\n{options_text}")
                value = prop.value
                hd.table_cell(render_value(value), **cell_style)
                hd.table_cell("", **cell_style)


def prop_tables(element_function_name, *args, **kwargs):
    """Renders multiple tables, each representing props for an element.
    Props are defined as a list of tuples, [group_name, [props]]
    Where the first entry is the name of the group
    (e.g. "Props inherited from StyledElement", Or just an empy string
    if they are the default props.)
    Each prop is a dict with name, type, default, and description attributes."""

    element = eval(f"hd.{element_function_name}")
    props_by_class = organize_props_by_class(
        element, element_function_name, *args, **kwargs
    )

    if len(props_by_class) == 0:
        return

    with hd.box(gap=8):
        hd.text("### Props")

        with hd.table():
            with hd.table_body():
                for i, (propclass, props) in enumerate(props_by_class):
                    with hd.scope(qualname(propclass)):
                        if len(props) == 0:
                            continue
                        if qualname(propclass) != qualname(element):
                            with hd.table_row():
                                with hd.table_cell(
                                    colspan=4,
                                    padding=(16, 8, 8, 0),
                                    border=None,
                                ):
                                    hd.text(
                                        f"#### Inherited from [{propclass.__name__}](/components/{propclass.__name__}):"
                                    )
                        header_row("Name", "Type", "Default", "Description")
                        prop_table(props)
