import hyperdiv as hd
from .text_utils import markdown_with_code_examples


def user_input():
    hd.text(
        """
        # Inputs

        Hyperdiv provides the following input components. These
        components can be used to gather input from the user in the
        form of text, clicks, and predetermined choice
        selection. Click each component to learn about its props and
        see examples.
        """
    )
    with hd.hbox(justify="space-between"):
        with hd.box():
            hd.text(
                """
                **Buttons**

                * [`button`](/components/button)
                """
            )

        with hd.box():
            hd.text(
                """
                **Text Inputs**

                * [`text_input`](/components/text_input)
                * [`password`](/components/password)
                * [`date`](/components/date)
                * [`search`](/components/search)
                """
            )

        with hd.box():
            hd.text(
                """
                **Choice Inputs**

                * [`checkbox`](/components/checkbox)
                * [`radio_group`](/components/radio_group)
                * [`select`](/components/select)
                * [`tabs`](/components/tabs)
                """
            )

        with hd.box():
            hd.text(
                """
                **Miscellaneous**

                * [`clipboard_button`](/components/clipboard_button)
                * [`dark_mode_toggle`](/components/dark_mode_toggle)
                """
            )

    markdown_with_code_examples(
        """
        ## Forms

        The input components above can be used individually in any way
        you want, in any part of the app. But often it's useful to
        group a bunch of inputs inside a
        [`form`](/components/form). Using forms, you can designate
        input components as *required* and "submit" the form only when
        it has been filled out properly.

        See [`form`](/components/form) for an example.

        ## Handling Clicks

        We'd like to shed some light on how input events,
        for now just clicks, are handled in Hyperdiv.

        All UI elements in Hyperdiv support the `clicked` UI input
        prop. `clicked` becomes `True` for a single frame, allowing
        your code to execute an action when the user clicks a UI
        component.

        Example:

        ```py
        if hd.button("Click Me").clicked:
            # Execute an "on click" action
            pass
        ```

        You could do the same with a `box`:
        ```py
        with hd.box(padding=16, width="fit-content", border=1) as box:
            hd.text("Click Me")
        if box.clicked:
            # Execute an "on click" action
            pass
        ```

        Any element that inherits from `InputElement` supports
        `clicked`, and that's pretty much all of them.

        ## Gotcha

        *TL;DR never render UI elements when handling a click.*

        Beware of writing code like this:

        ```py
        if hd.button("Click Me").clicked:
            hd.text("Clicked")
        ```

        The text `Clicked` will render for exactly one frame. On the
        next frame, `clicked` will be `False` and the text will no
        longer be rendered. You can verify this by clicking the button
        or box in the examples above, which causes a new frame to be
        rendered, and see that the text disappears.

        To persistently display the text after the first click, you
        would persist the click into a piece of state, and then render
        text based on the value of that state.

        ```py
        s = hd.state(was_clicked=False)
        if hd.button("Click Me").clicked:
            # Persist the state
            s.was_clicked = True
        if s.was_clicked:
            hd.text("Clicked")
        ```
        """
    )
