from ..core.element import Prop
from .styled_element import StyledElement
from .props import (
    OptionProp,
    BoolProp,
    OptionalStrProp,
    DirectionProp,
    AlignProp,
    JustifyProp,
    GapProp,
    WrapProp,
    AriaProp,
    GridTemplateAreasProp,
    GridTemplateRowsProp,
    GridRowProp,
)


class container(StyledElement):
    """
    `box` and `hbox` are the Hyperdiv container elements and the
    fundamental building blocks for creating rich, complex UIs. Any
    hyperdiv elements, including boxes, can be nested inside boxes.

    The elements nested inside a `box` will be stacked vertically. The
    elements nested inside an `hbox`, on the other hand, will be laid
    out side by side, horizontally.

    Boxes have a variety of props that can be used to control their
    dimensions, their look, and how their children are laid out.  The
    APIs for `box` and `hbox` are identical.

    `box` example:

    ```py
    with hd.box(gap=16):
        hd.button("First Child")
        hd.button("Second Child")
    ```

    `hbox` example:

    ```py
    with hd.hbox(gap=16):
        hd.button("First Child")
        hd.button("Second Child")
    ```

    Nesting example:

    ```py
    with hd.box(border="blue", gap=16, padding=16):
        hd.button("First Child")
        with hd.box(border="red", padding=16):
            hd.button("Nested Button")
    ```
    """

    is_container = True

    direction = DirectionProp("vertical")
    align = AlignProp("stretch")
    justify = JustifyProp("start")
    vertical_scroll = BoolProp(False)
    horizontal_scroll = BoolProp(False)
    gap = GapProp(None)
    layout = OptionProp("flex", ("flex", "grid"))
    # TODO: What's the type of this prop?
    grid_area = Prop(None)
    wrap = WrapProp(False)
    role = OptionalStrProp(None)
    aria = AriaProp(None)
    grid_template_columns = GridTemplateRowsProp(None)
    grid_template_rows = GridTemplateRowsProp(None)
    grid_template_areas = GridTemplateAreasProp(None)
    grid_row = GridRowProp(None)
    grid_column = GridRowProp(None)


box = container
vbox = container


class hbox(container):
    name = "container"
    direction = DirectionProp("horizontal")
