from .container import container
from .select_ import select
from .checkbox import checkbox
from .text_input import text_input
from .radio import radio

input_types = (
    text_input,
    select,
    radio,
    checkbox,
)


def get_form_inputs(children):
    inputs = []
    for c in children:
        if isinstance(c, input_types):
            inputs.append(c)
        elif c.is_container and not isinstance(c, form):
            inputs.extend(get_form_inputs(c.children))
    return inputs


class form(container):
    """The `form` component can be used to collect user input in bulk,
    using multiple input components, each of which can be designated
    as required or optional, and then perform an action on all the
    collected inputs after the form has been filled out properly.

    ```py
    with hd.box(gap=16):
        with hd.form(gap=16) as form:
            name = hd.text_input(
                "Your Name",
                required=True
            )
            food_choice = hd.radio_group(
                "Food Preference:",
                "No Preference",
                (
                    "No Preference",
                    "Fish",
                    "Chicken",
                    "Vegetarian"
                )
            )
            agree = hd.checkbox('Agree to terms', required=True)

        with hd.hbox(gap=8):
            submit = hd.button(
                "Submit",
                disabled=not form.can_submit
            )
            reset = hd.button("Reset")

        if submit.clicked:
            # Do something with name.value, food_choice.value
            pass
        if reset.clicked:
            form.reset()
    ```

    ### Gotcha

    Notice that you cannot access `form.can_submit` while inside the
    body of the form. The intuition for this is that until the form's
    body is completed, we cannot check the validity of its input
    components. In the example above, the form's `Submit` button is
    rendered outside the form's body, because it depends on
    `can_submit`.

    """

    is_container = True

    def __init__(self, *args, **kwargs):
        self._inputs = []
        self._inside_form = False
        super().__init__(self, *args, **kwargs)

    def __enter__(self):
        self._inputs = []
        self._inside_form = True
        return super().__enter__()

    def __exit__(self, *args):
        super().__exit__(*args)
        self._inputs = get_form_inputs(self.children)
        self._inside_form = False

    @property
    def can_submit(self):
        """[public] Whether all required fields have been filled out."""
        if self._inside_form:
            raise Exception("Cannot ask can_submit from inside the body of the form.")
        return all([i.valid and not (i.required and i.is_empty) for i in self._inputs])

    def reset(self):
        """[public] Reset all the inputs in the form to default values."""
        for i in self._inputs:
            i.reset()
