from ..core.element import Element
from .props import StrProp


class location(Element):
    """
    `location` is the building block for implementing application
    navigation and routing. It represents the value of the browser's
    location bar, and changing its props will change the browser's
    location bar.

    The main interface for modifying the location is the `go()`
    method, documented below.

    `location` has three props: `path`, `query_args`, and
    `hash_arg`. Suppose the browser's location bar displays the
    following location:

        http://localhost:8888/fruit?type=apple&family=jonathan#description

    Then `location`'s props and their values are as follows:

    * `path`: `"/fruit"`
    * `query_args`: `"type=apple&family=jonathan"`
    * `hash_arg`: `"description"`

    Here's how you'd implement routing based on the value of
    `hash_arg`. (We'll stick to `hash_arg` for this example, but
    implementing routing based on `path` works identically.)

    ```py
    loc = hd.location()

    with hd.box(gap=16):
      with hd.hbox(gap=8):
        # Render navigation buttons, which modify
        # the location when clicked:
        if hd.button("Home").clicked:
          loc.go(hash_arg="")
        if hd.button("Welcome").clicked:
          loc.go(hash_arg="welcome")
        if hd.button("Users").clicked:
          loc.go(hash_arg="users")

      # Render the page content, depending on location:
      if loc.hash_arg == "welcome":
          hd.text("## Welcome Page")
      elif loc.hash_arg == "users":
          hd.text("## Users Page")
      elif loc.hash_arg == "":
          hd.text("## Home Page")
    ```

    Notice the hash arg in the location bar changing when you click the
    navigation buttons. And if you reload the browser, the correct page
    is rendered still.

    """

    is_singleton = True

    path = StrProp("/")
    query_args = StrProp("")
    hash_arg = StrProp("")

    def go(self, path=None, query_args=None, hash_arg=""):
        """[public]

        The main interface to this component.

        Sets all three props at the same time. If you set `path`
        alone, the location `query_args` and `hash_arg` will be reset.

        Similarly, if `query_args` is set but `hash_arg` is not,
        `hash_arg` will be reset.

        Some examples:
        ```py
        # For each of the following examples, assume
        # the current location is: /foo?x=1#hello

                                New Location
                                ---------------
        go(path="/bar")      => "/bar"
        go(query_args="z=2") => "/foo?z=2"
        go(hash_arg="meow")  => "/foo?x=1#meow"
        ```

        """

        # TODO: Buggy

        if path is None:
            path = self.path
            if query_args is None:
                query_args = self.query_args
        else:
            query_args = ""
            hash_arg = ""

        if path == "":
            path = "/"

        self.path = path
        self.query_args = query_args
        self.hash_arg = hash_arg
