#
# Utility methods for parsing rectangular props
#


def parse_rectangle_style(
    value,
    value_type_check,
    value_to_str,
    names=("top", "right", "bottom", "left"),
    default=None,
):
    """
    A helper method that parses a single value (e.g. a number),
    or a dict that has 1-4 values
    corresponding to the sides/corners of a rectangle.
    The default value is used when an entry is ommitted
    from the dictionary.
    """
    # First check if it's a valid single value
    if value_type_check(value):
        return value_to_str(value)

    # Otherwise check if it's a dict with valid entries
    elif (
        isinstance(value, dict)
        and len(value) > 0
        and len(value) <= 4
        and all(value_type_check(value[x]) for x in value)
    ):
        str_values = dict()
        for name in names:
            if name in value:
                str_values[name] = value_to_str(value[name])
            else:
                str_values[name] = value_to_str(default)
        return str_values

    # Finally check if it's a list/tuple with valid entries
    elif (
        isinstance(value, (list, tuple))
        and len(value) > 0
        and len(value) <= 4
        and all(value_type_check(x) for x in value)
    ):
        str_values = dict()
        for index, name in enumerate(names):
            if index >= len(value):
                str_values[name] = value_to_str(default)
            else:
                str_values[name] = value_to_str(value[index])
        return str_values
    raise ValueError("Invalid prop format.")


#
# Color utils
#


def is_hex_color(color):
    if type(color) is not str:
        return False
    if not color.startswith("#"):
        return False

    # Remove the leading "#"" symbol
    color = color[1:]

    # Check if it's rgb, rrggbb, rrggbbaa
    if len(color) not in (3, 6, 8):
        return False

    # Check if it's actually hex
    try:
        int(color, 16)
        return True
    except ValueError:
        return False


def is_theme_color(color):
    if not isinstance(color, str):
        return False

    return theme_color_from_str(color) is not None


def theme_color_from_str(color):
    # Avoid cyclical imports
    from .theme import theme

    return getattr(theme(), color, None)


def is_color(color):
    return is_theme_color(color) or is_hex_color(color)


def parse_color(color):
    """
    Accepts the following types:
    * 1. None
    * 2. A hex color value #rrggbb, #rgb, or #rrggbbaa
    """
    if color is None:
        return None
    if is_hex_color(color):
        return color
    if is_theme_color(color):
        return theme_color_from_str(color)

    raise ValueError("Invalid color format.")


def to_rgba(color):
    color = parse_color(color)

    color = color[1:]
    r, g, b, a = ("0", "0", "0", "FF")
    if len(color) == 3:
        r = color[0] * 2
        g = color[1] * 2
        b = color[2] * 2
    elif len(color) >= 6:
        r = color[0:2]
        g = color[2:4]
        b = color[4:6]
    if len(color) == 8:
        a = color[6:8]

    return (int(r, 16), int(g, 16), int(b, 16), int(a, 16))


def is_light(rgba):
    r, b, g, a = rgba
    return (r + g + b) / 3 >= 128


def offset_color(rgba, offset):
    if rgba is None:
        return None
    else:
        r, g, b, a = rgba
        r = min(255, max(0, r + offset))
        g = min(255, max(0, g + offset))
        b = min(255, max(0, b + offset))

        return rgba_to_hex((r, g, b, a))


def rgba_to_hex(rgba):
    r, g, b, a = rgba
    return "#%02x%02x%02x%02x" % (r, g, b, a)


def highlight(color, amount=50):
    rgba = to_rgba(color)
    if is_light(rgba):
        return darken(rgba, amount=amount)
    else:
        return lighten(rgba, amount=amount)


def lighten(color, amount=50):
    if type(color) is str:
        color = to_rgba(color)
    return offset_color(color, amount)


def darken(color, amount=50):
    if type(color) is str:
        color = to_rgba(color)
    return offset_color(color, -amount)


#
# Misc utils
#


def is_non_negative_number(x):
    return type(x) in (int, float) and x >= 0


def number_to_str(value, units="px"):
    return f"{value}{units}"


def is_url(src):
    from urllib.parse import urlparse

    try:
        value = urlparse(src)
        # Require a domain and http(s).
        return value.netloc != "" and value.scheme in ("http", "https")
    except Exception:
        return False
