from ..core.element import Prop
from .prop_utils import parse_color, is_non_negative_number, number_to_str


class NativeProp(Prop):
    """
    A prop type that accepts all values of a given Python type.
    Suppose the prop is defined with:
    ```py
    NativeProp(int, 0)
    ```

    Then it will accept `int` values like `0`, `100`, `-10`, etc., and
    reject all other values.
    """

    def __init__(self, type_, *args, **kwargs):
        self.type_ = type_
        super().__init__(*args, **kwargs)

    def parse(self, value):
        if not isinstance(value, self.type_):
            raise ValueError(
                f"Value of type `{self.type_.__name__}` expected "
                f"but got value `{value}` of type `{type(value).__name__}`."
            )
        return value


class IntProp(NativeProp):
    """
    A prop type that accepts all values of Python type `int`.
    """

    def __init__(self, value, **kwargs):
        super().__init__(int, value, **kwargs)


class BoolProp(NativeProp):
    """A prop type that accepts all values of Python type `bool`, namely
    `True` and `False`.
    """

    def __init__(self, value, **kwargs):
        super().__init__(bool, value, **kwargs)


class StrProp(NativeProp):
    """A prop type that accepts all values of Python type `str`."""

    def __init__(self, value, **kwargs):
        super().__init__(str, value, **kwargs)


class PositiveIntProp(IntProp):
    """A prop type that accepts all *positive* values of Python type
    `int`, including `0`. Any negative `int` values like `-10`,
    `-100`, or values of any other type, will be rejected.
    """

    def parse(self, value):
        value = super().parse(value)
        if value < 0:
            raise ValueError(f"A positive int expected but got {value}")
        return value


class OptionalStrProp(StrProp):
    """
    A prop type that accepts all values of Python type `str`, but in
    addition it also accepts the value `None`.
    """

    def parse(self, value):
        if value is None:
            return value
        return super().parse(value)


class ListProp(Prop):
    """
    A prop type that accepts values of type `list` or `tuple`,
    regardless of what types of elements are stored in those lists
    or tuples.

    ```py
    # Example values:
    []
    [1, 2, 3]
    [1, "Hello", None]
    ```
    """

    def parse(self, value):
        if type(value) not in (list, tuple):
            raise ValueError("Value must be a list or a tuple.")
        return value


class TypedListProp(ListProp):
    """
    A prop type that accepts values of type `list` or `tuple`, where
    the elements stored in those lists are restricted to a specific
    type.

    Suppose a prop is defined as:

    ```py
    TypedListProp(
      int, # element type
      []   # default value (could be any list/tuple of ints)
    )
    ```

    Then:
    ```py
    # Good values:
    [],
    [1, 2, 3]
    (1, -10, 20)
    # Bad values:
    ["Hello"]
    [1, 2, "Steve"]
    ```
    """

    def __init__(self, type_, value, **kwargs):
        self.type_ = type_
        super().__init__(value, **kwargs)

    def parse(self, value):
        return [self.type_(elem) for elem in super().parse(value)]


class OptionProp(Prop):
    """
    A prop type that accepts values only from a given fixed collection
    of values.

    Suppose a prop is defined as:
    ```py
    prop = OptionProp(
      # Default value
      0,
      # Values to choose from
      (0, 1)
    )

    # Good values:
    0
    1
    # Examples of rejected values:
    -1
    2
    "Steve"
    False
    ```
    """

    def __init__(self, value, options, **kwargs):
        self.options = options
        super().__init__(value, **kwargs)

    def parse(self, value):
        if value not in self.options:
            options = ", ".join([str(o) for o in self.options])
            raise ValueError(f"Value {value} is not one of {options}")
        return value


class FontWeightProp(OptionProp):
    """
    Defines the font weight of the element. The only accepted values are:
    ```py
    None        # Undefined
    "normal"    # Normal font
    "bold"      # Bold font
    ```
    """

    def __init__(self, value, **kwargs):
        super().__init__(value, (None, "normal", "bold"), **kwargs)


class FontSizeProp(OptionProp):
    """
    Defines the font size of the element. The only accepted values are:
    ```py
    None
    "xx-small"
    "x-small"
    "small"
    "medium"
    "large"
    "x-large"
    "xx-large"
    "xxx-large"
    ```
    """

    def __init__(self, value, **kwargs):
        super().__init__(
            value,
            (
                None,
                "xx-small",
                "x-small",
                "small",
                "medium",
                "large",
                "x-large",
                "xx-large",
                "xxx-large",
            ),
            **kwargs,
        )


class FontsProp(Prop):
    """
    Defines a group of font faces.

    The accepted values are dictionaries mapping font names to font
    URLs or None. Example valid value:

    ```py
    {
       # Native font faces:
       "serif": None,
       "sans-serif": None,
       # This font is loaded from the web:
       "Nunito": "https://fonts.googleapis.com/css2?family=Nunito..."
    }
    ```
    """

    def parse(self, fonts):
        if fonts is None:
            return None

        # We only send webfonts that with urls
        webfonts = []
        for url in fonts.values():
            if url is not None and url not in webfonts:
                webfonts.append(url)
        return webfonts


class FontFamilyProp(Prop):
    """
    TODO
    """

    def parse(self, values):
        if values is None:
            return None

        # Checks if the font is included
        if type(values) in (list, tuple):
            return ", ".join(values)

        return values


class SizeProp(Prop):
    """
    Defines a size, such as the height or the width of an element.

    Accepted values:

    1. `None`
    2. A *positive* `int` or `float` number. This normally
       denotes a size in pixels.
    3. A percent string "X%", where 0 <= X <= 100.
    4. A viewport percent string "Xvh" or "Xvw", where 0 <= X <= 100.
       "Xvh" means X% of the vertical size (height) of the
       viewport. "Xvw" means X% of the horizontal size (width) of the
       viewport.

    Examples:
    ```py
    # Good values:
    None
    10
    500.5
    "30%"
    "78.5%"
    "10vh"
    "89.1vw"
    # Rejected values:
    -5
    "-80%"
    "1000vh"
    "140%"
    ```

    """

    def parse(self, size):
        def is_percent(value):
            if type(value) is str and value.endswith("%"):
                num = float(value[:-1])
                return num >= 0 and num <= 100

            return False

        def is_vwvh(value):
            if type(value) is str and (value.endswith("vw") or value.endswith("vh")):
                num = float(value[:-2])
                return num >= 0 and num <= 100

            return False

        if size is None:
            return None
        if is_non_negative_number(size):
            return number_to_str(size)
        elif is_percent(size) or is_vwvh(size):
            return size
        elif size == "fit-content":
            return size
        raise ValueError("Invalid size format.")


class ColorProp(Prop):
    """
    Represents a color value.

    Accepts the following types of values:

    1. `None`. This means "no color".
    2. A rgb/rgba hex color value in the format `#rgb` or `#rrggbb` or
    `#rrggbbaa`.
    3. A named color whose default value is taken from the `theme`
    element. These are:
        * `"black"`
        * `"white"`
        * `"blue"`
        * `"red"`
        * `"green"`
        * `"yellow"`
    """

    def parse(self, color):
        return parse_color(color)


class BorderProp(Prop):
    """
    Represents the border styling of the element, giving independent
    control over the four edges (top, right, bottom, left) making
    up the element's border.

    Accepts the following types of values:

    1. `None`. The element has no border.

    2. Tuple with three values: `(width, style, color)`. This border
       style is applied uniformly to all four edges.

        * `width`: An `int` representing the border thickness in pixels.
        * `style`: One of `"solid"`, `"dotted"`, `"dashed"`.
        * `color`: A color values as defined by [ColorProp](#ColorProp)

    3. An `int` value representing the border thickness in
       pixels. This border is applied uniformly to all four edges,
       with a default style and color.

    4. A valid `style` value as defined in (2). The style is applied
       to all four edges with default thickness and color.

    5. A dict with keys so that at least one of these keys defined: `"top"`,
       `"right"`, `"bottom"`, `"left"`, where the values of the keys
       are of type (1), (2), (3), or (4). Unspecified edges default to
       `None`.

    6. A list or tuple of size 4 or less, where each entry is of type
       (1), (2), (3), or (4). Each element of the list or tuple defines one
       edge of the border, starting at the top and going clockwise. If
       less than 4 edges are defined, the remaining edges default to
       `None`.

    Examples:
    ```py
    # To set all four edges simultaneously:
    None        # No border
    10          # Border with thickness 10 (default style, color)
    "dotted"    # Styled border (default thickness, color)
    "green"     # Colored border (default thickness, style)
    "#a2a4ff"   # Ditto

    # Defining thickness, style, and color at the same time,
    # for all four edges:
    (1, "solid", "red")
    (10, "dotted", "#abc")
    (10, "dotted", "#a1b2c3")
    (10, "dotted", "#a1b2c3")

    # To set edges selectively:
    # With dicts:
    dict(top=(1, "solid", "red"))
    dict(
      top=(2, "dotted", "green"),
      right="green",
      bottom="dotted",
      left=10
    )
    # With lists/tuples:
    dict([10])             # 10 pixel top border
    dict((
      10,                  # Top
      "green",             # Right
      (1, "solid", "red")  # Bottom
    ))
    # Above, the left border is implicitly None.
    ```
    """

    def parse(self, border):
        from .prop_utils import (
            parse_rectangle_style,
            is_color,
        )

        def is_border_style(style):
            return style in ["solid", "dotted", "dashed"]

        def border_to_str(border):
            """
            Converts a tuple/list (width, style, color) into a single string
            that will be valid on the frontend.
            """
            from .theme import theme

            if border is None:
                return None
            elif is_non_negative_number(border):
                thickness = border
                _, style, color = theme().input_border
                border = (thickness, style, color)
            elif is_color(border):
                color = parse_color(border)
                thickness, style, _ = theme().input_border
                border = (thickness, style, color)
            elif is_border_style(border):
                style = border
                thickness, _, color = theme().input_border
                border = (thickness, style, color)
            else:
                border = (border[0], border[1], parse_color(border[2]))

            return f"{number_to_str(border[0])} {border[1]} {border[2]}"

        def is_border(border):
            """
            Returns True if the border is None
            or a tuple/list with three entries (width, style, color)
            """
            return (
                border is None
                or (
                    isinstance(border, (list, tuple))
                    and len(border) == 3
                    and is_non_negative_number(border[0])
                    and is_border_style(border[1])
                    and is_color(border[2])
                )
                or is_color(border)
                or is_non_negative_number(border)
                or is_border_style(border)
            )

        try:
            return parse_rectangle_style(border, is_border, border_to_str, default=None)
        except ValueError:
            raise ValueError("Invalid border format.")


class OutlineProp(BorderProp):
    """
    TODO
    """

    def parse(self, outline):
        if outline is None or outline == "auto":
            return outline

        offset = 0
        # TODO: This is ambiguous, because a border could be
        # [1,1,1,1], which case this code takes the border as (1,1,1)
        # ans the offset as 1. It also does not do enough checking on
        # the args. The 4th arg could be a string, tuple, etc.
        if type(outline) in (list, tuple) and len(outline) == 4:
            offset = outline[3]
            outline = outline[:-1]

        return dict(outline=super().parse(outline), offset=number_to_str(offset))


class BorderRadiusProp(Prop):
    """
    Defines the border radius (aka corner radius) of an element. It
    gives independent control over the radii of all four corners of
    the border.

    Accepts the following types of values:

    1. A positive number (int or float). This sets the radius of all
       four corners simultaneously.

    2. A dict of size 4 or less, with valid keys `"topleft"`,
       `"topright"`, `"bottomleft"`, `"bottomright"`, and *positive*
       number values as defined in (1). This dict defines the radii of
       the four corners independently.

    3. A list or tuple of size 4 or less, containing positive numbers
       as defined in (1). The list represents corner radii starting at
       the `"topleft"` corner and going clockwise.

    The values of missing corners from (2) and (3) default to 0.

    Examples:
    ```py
    0
    20.3
    dict(topleft=10)
    dict(topleft=20, bottomright=1.1)
    (10,)        # Sets only the top-left corner
    (10, 20)     # Bottom corners default to 0
    [10, 20, 30] # Bottom-left corner defaults to 0
    ```
    """

    def parse(self, radius):
        from .prop_utils import parse_rectangle_style

        try:
            return parse_rectangle_style(
                radius,
                is_non_negative_number,
                lambda x: x,
                names=("topleft", "topright", "bottomright", "bottomleft"),
                default=0,
            )
        except ValueError:
            raise ValueError("Invalid border radius format.")


class PaddingProp(Prop):
    """
    Defines the padding of the element, meaning the amount of space
    between the edges and the inner content. It gives independent
    padding control over the four edges/sides of the element.

    Accepts the following types of values:

    1. A positive number (int or float), representing the amount of
       padding. This sets the padding of all four edges
       simultaneously.

    2. A dict of size 4 or less, with valid keys `"top"`, `"right"`,
       `"bottom"`, `"left"`, and *positive* number values as defined
       in (1). This dict defines the padding values of the four
       edges independently.

    3. A list or tuple of size 4 or less, containing positive numbers
       as defined in (1). The list represents the padding values of
       the four edges, starting at the `"top"` edge and going
       clockwise.

    The values of missing edges from (2) and (3) default to 0.

    Examples:
    ```py
    0
    20.3
    dict(top=10)
    dict(top=20, right=1.1)
    (10,)        # Sets only the top edge
    (10, 20)     # Bottom and left edges default to 0 padding
    [10, 20, 30] # Left edge defaults to 0
    ```

    """

    def parse(self, padding):
        from .prop_utils import parse_rectangle_style

        try:
            return parse_rectangle_style(
                padding, is_non_negative_number, lambda x: x, default=0
            )
        except ValueError:
            raise ValueError("Invalid padding format.")


class CursorProp(OptionProp):
    """
    Describes the mouse cursor when the mouse is hovering over the element.

    The accepted values are:

    * `None`. The browser's default.
    * `"auto"`. Equivalent to `None`.
    * `"pointer"`. Hand icon indicating that the element is clickable.

    """

    def __init__(self, value, **kwargs):
        super().__init__(value, (None, "auto", "pointer"), **kwargs)


class ThemeModeProp(OptionProp):
    """
    Controls the theme's dark/light mode setting.

    Accepted values:

    * `"system"`. The light/dark mode follows the system setting.
    * `"dark"`. Dark mode.
    * `"light"`. Light mode.
    """

    def __init__(self, value, **kwargs):
        super().__init__(value, ("system", "dark", "light"), **kwargs)


class SystemModeProp(OptionProp):
    """
    (Read-only) Describes which value the system dark/light mode
    setting is currently set to.

    Possible values:

    * `"dark"`. The system setting is currently set to dark mode.
    * `"light"`. The system setting is currently set to light mode.

    """

    def __init__(self, value, **kwargs):
        super().__init__(value, ("dark", "light"), **kwargs)


alignments = ("start", "end", "center", "stretch")


class AlignProp(OptionProp):
    """
    Defines how the container's children are aligned on the axis
    *opposite* to the container's direction.

    See [JustifyProp](#JustifyProp) for this prop's complement.

    The possible values are `"start"`, `"end"`, `"center"`, and `"stretch"`.

    For containers whose direction is set to `"vertical"` (for
    example, `box`), this prop determines how children are aligned on
    the *horizontal* axis. For `"horizontal"` containers (for example,
    `hbox`), this prop determines how children are aligned on the
    *vertical* axis.

    For vertical containers:

    * `"start"`. The children are left-aligned.
    * `"end"`. The children are right-aligned.
    * `"center"`. The children are center-aligned horizontally.
    * `"stretch"`. The children are horizontally stretched to fill the
      space. This setting will not affect children with fixed widths.

    For horizontal containers:

    * `"start"`. The children are top-aligned.
    * `"end"`. The children are bottom-aligned.
    * `"center"`. The children are center-aligned vertically.
    * `"stretch"`. The children are vertically stretched to fill the
      space. This setting will not affect children with fixed heights.

    """

    def __init__(self, value, **kwargs):
        super().__init__(value, alignments, **kwargs)


directions = ("vertical", "horizontal")


class DirectionProp(OptionProp):
    """
    Describes in which direction the container element's content is
    rendered.

    Values:

    * `"vertical"`. The container's children are rendered vertically,
      meaning each next element is *under* the previous.

    * `"horizontal"`. The container's children are rendered
      horizontally, meaning each next child is to the right of the
      previous.
    """

    def __init__(self, value, **kwargs):
        super().__init__(value, directions, **kwargs)


justifies = (
    "start",
    "end",
    "center",
    "space-between",
    "space-around",
    "space-evenly",
)


class JustifyProp(OptionProp):
    """
    Defines how the container's children are aligned on the *same*
    axis as the container's direction.

    See [AlignProp](#AlignProp) for this prop's complement.

    The possible values are:

    * `"start"`
    * `"end"`
    * `"center"`
    * `"space-between"`
    * `"space-around"`
    * `"space-evenly"`

    For horizontal containers such as `hbox`:

    * `"start"`. The children are left-aligned.

    * `"end"`. The children are right-aligned.

    * `"center"`. The children are centered horizontally.

    * `"space-between"`. The children are equally spaced out
      horizontally, taking up the full width of the container. Space
      is added *between* the children, but not between the left edge
      and the 1st child, or between the right edge and the last
      child.

    * `"space-around"`. As above, the children are equally spaced out
      horizontally. Equal space is added *around* the children, so the
      space between the left edge and the 1st child, or the space
      between the right edge and the last child, will be equal to half
      of the space between children.

    * `"space-evenly"`. As above, the children are equally spaced out
      horizontally. Space is added such that the space between the
      left edge and the 1st child, the space between the right edge
      and the last child, and the space between children, are equal.

    For vertical containers such as `box`:

    * `"start"`. The children are top-aligned.

    * `"end"`. The children are bottom-aligned.

    * `"center"`. The children are centered vertically.

    * `"space-between"`. The children are equally spaced out
      vertically, taking up the full height of the container. Space
      is added *between* the children, but not between the top edge
      and the 1st child, or between the bottom edge and the last
      child.

    * `"space-around"`. As above, the children are equally spaced out
      vertically. Equal space is added *around* the children, so the
      space between the top edge and the 1st child, or the space
      between the bottom edge and the last child, will be equal to half
      of the space between children.

    * `"space-evenly"`. As above, the children are equally spaced out
      vertically. Space is added such that the space between the
      top edge and the 1st child, the space between the bottom edge
      and the last child, and the space between children, are equal.
    """

    def __init__(self, value, **kwargs):
        super().__init__(value, justifies, **kwargs)


wraps = ("nowrap", "wrap", "reverse")


class WrapProp(Prop):
    """
    Determines if the children of a container element should wrap when
    they run out of space.

    In a vertical container, this setting determines if the children
    should wrap, spilling into a newly created layout "column", when
    the height of the container is too small to fit all the children.

    In a horizontal container, the children spill out into a newly
    created layout "row", when the width of the container is too small
    to fit all the children.

    Values:

    * `True` or `"wrap"`: The children should wrap.
    * `False` or `"nowrap"`: The children should not wrap.
    """

    def parse(self, wrap):
        if wrap is True:
            return "wrap"
        if wrap is False:
            return "nowrap"
        if wrap in wraps:
            return wrap
        raise ValueError("Invalid wrap value.")


class AriaProp(Prop):
    """TODO"""

    def parse(self, aria):
        if aria is None:
            return None
        if not isinstance(aria, dict):
            raise ValueError("aria must be a dict.")
        for k, v in aria.items():
            if not (isinstance(k, str) and k.isalpha()):
                raise ValueError("aria keys must be alphabetic strings.")
            if not isinstance(v, (str, int, float, bool)):
                raise ValueError("aria values must be primitive.")
        return aria


class GapProp(SizeProp):
    """
    Determines the size of the gap between a container's children.

    It takes the same types of values as [SizeProp](#SizeProp).
    """

    def parse(self, gap):
        if isinstance(gap, (list, tuple)):
            if len(gap) != 2:
                raise ValueError("Invalid gap value.")
            return f"{super().parse(gap[0])} {super().parse(gap[1])}"
        else:
            return super().parse(gap)


class GridRowProp(Prop):
    """
    TODO
    """

    def parse(self, gridrow):
        if gridrow is None:
            return None
        if (
            type(gridrow) in (list, tuple)
            and len(gridrow) == 2
            and all([isinstance(x, int) for x in gridrow])
        ):
            return f"{gridrow[0]} / {gridrow[1]}"

        raise ValueError(
            "Invalid grid row / column property. Must be a tuple of length 2."
        )


class GridTemplateAreasProp(Prop):
    """
    TODO
    """

    def parse(self, areas):
        if areas is None:
            return None
        # TODO: Should this be a list of strings, or convert the
        # return list to contain only strings?
        elif type(areas) in (tuple, list):
            return areas

        raise ValueError("Invalid grid_template_areas format.")


class GridTemplateRowsProp(Prop):
    """
    TODO
    """

    def parse(self, rows):
        # TODO: Write tests for this code
        def is_fr(value):
            if type(value) is str and (value.endswith("fr")):
                num = float(value[:-2])
                return num >= 0 and num <= 100

            return False

        if rows is None:
            return None
        elif type(rows) in (tuple, list):
            rowstr = ""
            for i in rows:
                if type(i) is int:
                    if i < 0:
                        raise ValueError("Value must be a positive int.")
                elif i != "auto" and not is_fr(i):
                    raise ValueError("Invalid grid_template_rows / columns.")
                rowstr += f"{i} "

            return rowstr.strip()

        raise ValueError("Invalid grid_template_rows / columns.")


class SrcProp(Prop):
    """
    Represents the location of a local or a remote resource, such as an
    image.

    If it's a local resource, it should be a path on the filesystem,
    pointing to a file that exists.

    If it's a remote resource, it should be a URL.

    Example values:
    ```py
    # Local path
    "/foo/bar/hello.jpg"
    # Remote URL
    "https://domain.com/foo/bar/image.png"
    ```
    """

    def parse(self, src):
        import os
        import xxhash

        from ..core.utils import static_directory
        from .prop_utils import is_url

        if src is None or src == "":
            return src

        if is_url(src):
            return src

        src = os.path.abspath(src)
        if not os.path.exists(src):
            raise ValueError("File does not exist.")

        sourcedir = os.path.dirname(src)
        filename = os.path.basename(src)

        if sourcedir != static_directory:
            sourcedir_hash = xxhash.xxh32(sourcedir).hexdigest()
            filename = f"{sourcedir_hash}_{filename}"

        hosted_path = os.path.join(static_directory, filename)
        # Do not copy the file if it already
        # exists at the destination
        if not os.path.exists(hosted_path):
            os.symlink(src, hosted_path)

        return f"/static/{filename}"
