from ..core.element import Element


class slot(Element):
    """
    A slot is an invisible placeholder in the document that can be
    filled with content at a later point. The interface for filling a
    slot is similar to filling a `box`, using the Python `with`
    statement:

    ```py
    # Create a slot
    slot = hd.slot()

    # Fill the slot
    with slot:
        hd.text("Slot content.")
        with hd.box(border=1, padding=8):
            hd.text('''
                A slot can be filled with
                arbitrary components.
            ''')
    ```

    Once filled, slots disappear from the document, and so cannot be
    re-filled. Attempting to re-fill an already filled slot will raise
    an error.

    Slots will not be needed for most use cases, but they may come in
    handy in niche situations. Here's an example pattern where we
    create a "page" abstraction, but we leave its body as a slot to be
    filled by the user of the page abstraction:

    ```py
    class Page:
        def __init__(self):
            with hd.box(border=1, padding=16, gap=8):
                hd.text("## Title")
                self.body_slot = hd.slot()
                hd.text("Footer")

        @property
        def body(self):
            return self.body_slot

    p = Page()
    with p.body:
        hd.text("Hello")
        if hd.checkbox("Check it").checked:
            hd.text("You checked it.")
    ```

    """

    is_container = True

    def __init__(self, key=None):
        super().__init__(key=key)
        self._collector = self._context.collector_stack.current()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        index = self._collector_index
        if index < 0:
            raise Exception("Cannot re-enter slot.")
        self._collector[index : index + 1] = self.children

    @property
    def _collector_index(self):
        try:
            return self._collector.index(self)
        except ValueError:
            return -1

    @property
    def is_empty(self):
        return self._collector_index >= 0
