from ..core.element import RemoveProp
from ..elements.theme import theme_prop
from .input_element import InputElement
from .text import text
from .container import container
from .props import SizeProp, SizeProp, PositiveIntProp


class table_cell(container):
    """
    `table_cell` represents a cell in a
    [table](/components/table). `table_cell` is a subclass of
    [box](/components/box) and so it accepts the same props as `box`,
    except `grow`, `shrink`, and `grid_area`. Since it's a `box`, it
    can have arbitary Hyperdiv components inside it.

    In addition to `box` props, it accepts the following props:

    * `colspan`: The number of cells this cell should span toward the
      right.
    * `rowspan`: The number of rows this cell should span
      downward.

    """

    # Props that the superclass supports but this class doesn't:
    grow = RemoveProp()
    shrink = RemoveProp()
    grid_area = RemoveProp()
    # Overridden prop:
    border = theme_prop("container_border")
    width = SizeProp("100%")
    height = SizeProp("100%")
    # New props:
    colspan = PositiveIntProp(1)
    rowspan = PositiveIntProp(1)

    def on_render(self, *label, **kwargs):
        # Render the label, if any, inside the cell
        if label:
            with self:
                text(*label)


class table_row(InputElement):
    """
    A `table_row` is a container that represents a row in a
    [table](/components/table). `table_row` accepts as children a
    sequence of [table_cell](/components/table_cell)s, and will be
    rendered in the order given. Any other types of children will be
    ignored.
    """

    is_container = True


class table_body(InputElement):
    """
    A `table_head` represents the body section of a
    table. `table_head` is a container that accepts zero or more
    [table_row](/components/table_row)s as children. Any other types
    of children are ignored.
    """

    is_container = True


class table_head(InputElement):
    """
    A `table_head` represents the header section of a
    table.

    `table_head` is a container that accepts zero or more
    [table_row](/components/table_row)s as children. Any other types
    of children are ignored.

    Typically, there is a single row, and it contains as children the
    [table_cell](/components/table_cell)s that contain the column
    headers of the table.

    """

    is_container = True
    background_color = theme_prop("surface_color")


class table_foot(InputElement):
    """
    A `table_foot` represents the footer section of a
    table. `table_foot` is a container that accepts zero or more
    [table_row](/components/table_row)s as children. Any other types
    of children are ignored.
    """

    is_container = True
    background_color = theme_prop("surface_color")


class table(InputElement):
    """
    The top-level component for creating tables.

    See the other components involved:

    * [table_head](/components/table_head)
    * [table_foot](/components/table_foot)
    * [table_row](/components/table_row)
    * [table_cell](/components/table_cell)

    `table` is a container that accepts as children a `table_head`,
    followed by a `table_body`, followed by a `table_foot`, but any of
    those components can be missing.

    Here's a complete table example. Notice that
    [table_cell](/components/table_cell) accepts the same arguments as
    [box](/components/box), so you can nest arbitrary Hyperdiv
    components in table cells.

    ```py
    with hd.table():
        with hd.table_head():
            with hd.table_row():
                hd.table_cell("Users", colspan=4, padding=8)
            with hd.table_row():
                hd.table_cell('ID', padding=8)
                hd.table_cell('Name', padding=8)
                hd.table_cell('Age', padding=8)
                hd.table_cell('')
        with hd.table_body():
            with hd.table_row():
                hd.table_cell(0, padding=8, justify="center")
                hd.table_cell("Neil", padding=8, justify="center")
                hd.table_cell(32, padding=8, justify="center")
                with hd.table_cell(
                    direction="horizontal",
                    padding=8,
                    gap=8,
                    justify="center"
                ):
                    hd.icon_button(hd.Icons.PenSolid)
                    hd.icon_button(hd.Icons.CircleXmark)
            with hd.table_row():
                hd.table_cell(1, padding=8, justify="center")
                hd.table_cell("Angie", padding=8, justify="center")
                hd.table_cell(25, padding=8, justify="center")
                with hd.table_cell(
                    direction="horizontal",
                    padding=8,
                    gap=8,
                    justify="center"
                ):
                    hd.icon_button(hd.Icons.PenSolid)
                    hd.icon_button(hd.Icons.CircleXmark)
        with hd.table_foot():
            with hd.table_row():
                hd.table_cell("Hello", colspan=4, padding=8)
    ```

    """

    is_container = True
