from nose.tools import assert_raises
from ...core.unittests import simple_test
from ..props import (
    IntProp,
    BoolProp,
    StrProp,
    PositiveIntProp,
    OptionalStrProp,
    ListProp,
    TypedListProp,
    AriaProp,
    BorderProp,
    BorderRadiusProp,
    ColorProp,
    CursorProp,
    FontFamilyProp,
    FontSizeProp,
    FontWeightProp,
    PaddingProp,
    SizeProp,
    WrapProp,
    OutlineProp,
    GapProp,
    GridRowProp,
    GridTemplateAreasProp,
    GridTemplateRowsProp,
    SrcProp,
)


def test_int_prop():
    prop = IntProp(None)
    assert prop.parse(10) == 10
    # Maybe weird, but True is an instance of int, and is equivalent
    # to 1. False is equivalent to 0.
    assert prop.parse(True) is True
    assert prop.parse(True) == 1
    assert_raises(ValueError, prop.parse, 10.1)
    assert_raises(ValueError, prop.parse, "foo")
    assert_raises(ValueError, prop.parse, None)


def test_bool_prop():
    prop = BoolProp(None)
    assert prop.parse(False) is False
    assert prop.parse(True) is True
    assert_raises(ValueError, prop.parse, 9)
    assert_raises(ValueError, prop.parse, None)
    assert_raises(ValueError, prop.parse, "Hello")


def test_str_prop():
    prop = StrProp(None)
    assert prop.parse("Hello") == "Hello"
    assert prop.parse("") == ""
    assert_raises(ValueError, prop.parse, 9)
    assert_raises(ValueError, prop.parse, None)
    assert_raises(ValueError, prop.parse, True)
    assert_raises(ValueError, prop.parse, object())


def test_optional_str_prop():
    prop = OptionalStrProp(None)
    assert prop.parse("Hello") == "Hello"
    assert prop.parse("") == ""
    assert prop.parse(None) is None
    assert_raises(ValueError, prop.parse, 9)
    assert_raises(ValueError, prop.parse, True)
    assert_raises(ValueError, prop.parse, object())


def test_positive_int_prop():
    prop = PositiveIntProp(None)
    assert prop.parse(0) == 0
    assert prop.parse(10) == 10
    assert prop.parse(True) == 1
    assert prop.parse(False) == 0
    assert_raises(ValueError, prop.parse, -1)
    assert_raises(ValueError, prop.parse, None)
    assert_raises(ValueError, prop.parse, "Hello")


def test_list_prop():
    prop = ListProp(None)
    assert prop.parse([]) == []
    assert prop.parse([1, 2, "Hello"]) == [1, 2, "Hello"]
    assert_raises(ValueError, prop.parse, 10)
    assert_raises(ValueError, prop.parse, None)
    assert_raises(ValueError, prop.parse, "Hello")
    assert_raises(ValueError, prop.parse, False)


def test_typed_list_prop():
    # Note: This type converts the elements to the given type, it
    # doesn't fail if the type doesn't match.
    prop = TypedListProp(int, None)
    assert prop.parse([]) == []
    assert prop.parse([1, 2, 3]) == [1, 2, 3]
    assert prop.parse([1, 2, "3"]) == [1, 2, 3]
    assert_raises(ValueError, prop.parse, 10)
    assert_raises(ValueError, prop.parse, None)
    assert_raises(ValueError, prop.parse, "Hello")
    assert_raises(ValueError, prop.parse, False)
    assert_raises(ValueError, prop.parse, ["foo"])
    assert_raises(ValueError, prop.parse, [1, "foo"])

    prop = TypedListProp(str, None)
    assert prop.parse([]) == []
    assert prop.parse(["1", "2", "3"]) == ["1", "2", "3"]
    assert prop.parse([1, 2]) == ["1", "2"]
    assert prop.parse([True]) == ["True"]
    assert_raises(ValueError, prop.parse, 10)
    assert_raises(ValueError, prop.parse, None)
    assert_raises(ValueError, prop.parse, "Hello")
    assert_raises(ValueError, prop.parse, False)


def test_aria_prop():
    prop = AriaProp(None)
    d = dict(a=1, b="2", c=True, d=5.6)
    assert prop.parse(d) == d
    assert prop.parse(None) is None
    assert_raises(ValueError, prop.parse, 1)
    assert_raises(ValueError, prop.parse, "Hello")
    assert_raises(ValueError, prop.parse, {1: "hello"})
    assert_raises(ValueError, prop.parse, {object(): "hello"})
    assert_raises(ValueError, prop.parse, {"x": dict()})
    assert_raises(ValueError, prop.parse, {"1x": "Hello"})
    assert_raises(ValueError, prop.parse, {"&": "Hello"})


@simple_test
def test_border_prop(ctx):
    import hyperdiv as hd

    prop = BorderProp(None)
    border = (1, "solid", "#eee")
    border_str = "1px solid #eee"
    assert prop.parse(border) == border_str
    assert prop.parse(None) is None
    assert prop.parse([border]) == dict(
        top=border_str, bottom=None, left=None, right=None
    )
    assert prop.parse([border, border]) == dict(
        top=border_str, bottom=None, left=None, right=border_str
    )
    assert prop.parse([border, None, border]) == dict(
        top=border_str, bottom=border_str, left=None, right=None
    )
    assert prop.parse([border, None, border, border]) == dict(
        top=border_str, bottom=border_str, left=border_str, right=None
    )

    width, style, color = hd.theme().input_border
    assert prop.parse(10) == f"10px {style} {color}"
    assert prop.parse("#a1b2c1") == f"{width}px {style} #a1b2c1"
    assert prop.parse("dotted") == f"{width}px dotted {color}"
    assert prop.parse([10]) == dict(
        top=f"10px {style} {color}", left=None, bottom=None, right=None
    )

    assert prop.parse([10, (1, "solid", "red")]) == dict(
        top=f"10px {style} {color}",
        right=f"1px solid {hd.theme().red}",
        bottom=None,
        left=None,
    )

    assert prop.parse(dict(top=10, right=(1, "solid", "red"))) == dict(
        top=f"10px {style} {color}",
        right=f"1px solid {hd.theme().red}",
        bottom=None,
        left=None,
    )

    assert_raises(ValueError, prop.parse, "Stuff")
    assert_raises(
        ValueError,
        prop.parse,
        [border, border, border, border, border, border],
    )
    assert_raises(ValueError, prop.parse, [])
    assert_raises(ValueError, prop.parse, ["Hello", object()])


@simple_test
def test_outline_prop(ctx):
    import hyperdiv as hd

    prop = OutlineProp(None)
    border = (1, "solid", "#eee")
    border_str = "1px solid #eee"

    assert prop.parse(border) == dict(outline=border_str, offset="0px")
    assert prop.parse(None) is None
    assert prop.parse([border]) == dict(
        outline=dict(top=border_str, bottom=None, left=None, right=None), offset="0px"
    )
    assert prop.parse([border, border]) == dict(
        outline=dict(top=border_str, bottom=None, left=None, right=border_str),
        offset="0px",
    )
    assert prop.parse([border, None, border]) == dict(
        outline=dict(top=border_str, bottom=border_str, left=None, right=None),
        offset="0px",
    )

    width, style, color = hd.theme().input_border
    assert prop.parse(10) == dict(outline=f"10px {style} {color}", offset="0px")
    assert prop.parse("#a1b2c1") == dict(
        outline=f"{width}px {style} #a1b2c1", offset="0px"
    )
    assert prop.parse("dotted") == dict(
        outline=f"{width}px dotted {color}", offset="0px"
    )

    assert prop.parse((1, "solid", "#a1b1c1", 5)) == dict(
        outline="1px solid #a1b1c1", offset="5px"
    )

    assert_raises(ValueError, prop.parse, "Stuff")
    assert_raises(
        ValueError,
        prop.parse,
        [border, border, border, border, border, border],
    )
    assert_raises(ValueError, prop.parse, [])
    assert_raises(ValueError, prop.parse, ["Hello", object()])


def test_gap_prop():
    prop = GapProp(None)
    assert prop.parse(1) == "1px"
    assert prop.parse(None) is None
    assert prop.parse("1%") == "1%"
    assert prop.parse((1, 2)) == "1px 2px"
    assert prop.parse(("1%", "2vh")) == "1% 2vh"
    assert_raises(ValueError, prop.parse, ("1%", 2, 3, 5))
    assert_raises(ValueError, prop.parse, "Foo")


def test_grid_row_prop():
    prop = GridRowProp(None)
    assert prop.parse(None) is None
    assert prop.parse((1, 2)) == "1 / 2"
    assert_raises(ValueError, prop.parse, ("Stuff", "Bunnies"))
    assert_raises(ValueError, prop.parse, (1, 2, 3))


def test_grid_template_areas_prop():
    prop = GridTemplateAreasProp(None)
    assert prop.parse(None) is None
    assert prop.parse((1, 2, 3)) == (1, 2, 3)
    assert prop.parse(("1", "2", "3", "4")) == ("1", "2", "3", "4")
    assert_raises(ValueError, prop.parse, "Foo")
    assert_raises(ValueError, prop.parse, 10)
    assert_raises(ValueError, prop.parse, dict(x=1))


def test_grid_template_rows_prop():
    prop = GridTemplateRowsProp(None)
    assert prop.parse(None) is None
    assert prop.parse((1, 2, 3)) == "1 2 3"
    assert prop.parse(("1fr", 2, 3, "auto")) == ("1fr 2 3 auto")
    assert_raises(ValueError, prop.parse, ("hello",))
    assert_raises(ValueError, prop.parse, (0.4,))
    assert_raises(ValueError, prop.parse, (-1,))
    assert_raises(ValueError, prop.parse, 10)
    assert_raises(ValueError, prop.parse, dict(x=1))


def test_border_radius_prop():
    prop = BorderRadiusProp(None)
    assert prop.parse(1) == 1
    assert prop.parse([1]) == dict(topleft=1, topright=0, bottomright=0, bottomleft=0)
    assert prop.parse([1, 2]) == dict(
        topleft=1,
        topright=2,
        bottomright=0,
        bottomleft=0,
    )
    assert prop.parse([1, 2, 3]) == dict(
        topleft=1, topright=2, bottomright=3, bottomleft=0
    )
    assert prop.parse([1, 2, 3, 4]) == dict(
        topleft=1, topright=2, bottomright=3, bottomleft=4
    )

    assert_raises(ValueError, prop.parse, "Hello")
    assert_raises(ValueError, prop.parse, ["Bunnies", "Squirrels"])
    assert_raises(ValueError, prop.parse, dict(random_stuff="Bunnies"))
    assert_raises(ValueError, prop.parse, -1)


@simple_test
def test_color_prop(ctx):
    import hyperdiv as hd

    prop = ColorProp(None)

    assert prop.parse("#fff") == "#fff"
    assert prop.parse("#fffeee") == "#fffeee"
    assert prop.parse("#fffeeeaa") == "#fffeeeaa"
    assert prop.parse("red") == hd.theme().red
    assert prop.parse(None) is None

    assert_raises(ValueError, prop.parse, "#ffeefff")
    assert_raises(ValueError, prop.parse, "ffeeff")
    assert_raises(ValueError, prop.parse, "#ffzzff")
    assert_raises(ValueError, prop.parse, "hello")
    assert_raises(ValueError, prop.parse, 1)


def test_cursor_prop():
    prop = CursorProp(None)
    assert prop.parse("auto") == "auto"
    assert prop.parse("pointer") == "pointer"
    assert prop.parse(None) is None
    assert_raises(ValueError, prop.parse, "hello")
    assert_raises(ValueError, prop.parse, 1)


def test_font_family_prop():
    prop = FontFamilyProp(None)
    assert prop.parse(None) is None
    assert prop.parse(("helvetica", "sans-serif")) == "helvetica, sans-serif"
    assert prop.parse("helvetica") == "helvetica"


def test_font_size_prop():
    prop = FontSizeProp(None)
    assert prop.parse("medium") == "medium"
    assert_raises(ValueError, prop.parse, 10)


def test_font_weight_prop():
    prop = FontWeightProp(None)
    assert prop.parse("bold") == "bold"
    assert_raises(ValueError, prop.parse, "blue")


def test_padding_prop():
    prop = PaddingProp(None)
    assert prop.parse(1) == 1
    assert prop.parse([1]) == dict(top=1, bottom=0, left=0, right=0)
    assert prop.parse([1, 2]) == dict(top=1, bottom=0, left=0, right=2)
    assert prop.parse([1, 2, 3]) == dict(top=1, bottom=3, left=0, right=2)
    assert prop.parse([1, 2, 3, 4]) == dict(top=1, bottom=3, left=4, right=2)
    assert prop.parse(dict(top=3, bottom=2)) == dict(top=3, bottom=2, left=0, right=0)
    assert_raises(ValueError, prop.parse, "Hello")
    assert_raises(ValueError, prop.parse, ["Bunnies", "Squirrels"])
    assert_raises(ValueError, prop.parse, dict(random_stuff="Bunnies"))
    assert_raises(ValueError, prop.parse, -1)


def test_size_prop():
    prop = SizeProp(None)
    assert prop.parse(1) == "1px"
    assert prop.parse("50%") == "50%"
    assert prop.parse("10vw") == "10vw"
    assert prop.parse(None) is None
    assert_raises(ValueError, prop.parse, "Hello")
    assert_raises(ValueError, prop.parse, -1)
    assert_raises(ValueError, prop.parse, dict(width="50px"))


def test_wrap_prop():
    prop = WrapProp(None)
    assert prop.parse(True) == prop.parse("wrap") == "wrap"
    assert prop.parse(False) == prop.parse("nowrap") == "nowrap"
    assert_raises(ValueError, prop.parse, "something")


def test_src_prop():
    import sys

    prop = SrcProp(None)
    assert prop.parse(None) is None
    assert prop.parse("") == ""
    assert prop.parse("http://google.com") == "http://google.com"
    assert_raises(ValueError, prop.parse, "Foo")

    # TODO: test on an actual file.
    prop.parse(sys.argv[0])
