from ..core.element import Element
from .props import (
    SystemModeProp,
    ThemeModeProp,
    ColorProp,
    FontFamilyProp,
    BorderRadiusProp,
    PaddingProp,
    BorderProp,
)


def light_or_dark(propname):
    def getter(theme_element):
        """Returns the corresponding theme color for
        depending on whether the current theme is light or
        dark mode.
        """
        mode = "dark" if theme_element.is_dark else "light"
        return getattr(theme_element, f"{mode}_{propname}")

    return getter


class theme(Element):
    is_singleton = True
    is_collected = False

    # Dark / light mode settings
    # TODO: only assignable from the frontend.
    system_mode = SystemModeProp("dark")
    mode = ThemeModeProp("system")

    # These remain the same in both light and dark mode.
    blue = ColorProp("#4A5AD3")
    green = ColorProp("#00999C")
    yellow = ColorProp("#F4BD59")
    red = ColorProp("#EE6869")
    black = ColorProp("#000000")
    white = ColorProp("#ffffff")

    # Aliases for the above
    primary_color = blue
    error_color = red
    warning_color = yellow
    success_color = green

    font_family = FontFamilyProp("Nunito")
    border_radius = BorderRadiusProp(0)
    input_padding = PaddingProp(8)

    # Dark mode values
    dark_background_color = ColorProp("#030405")
    dark_surface_color = ColorProp("#2C2C2D")
    dark_font_color = ColorProp("#fff")
    dark_container_border = BorderProp((1, "solid", "#545555"))
    dark_input_border = BorderProp((1, "solid", "#545555"))
    dark_modal_cover_color = ColorProp("#ffffff18")

    # Light mode values
    light_background_color = ColorProp("#fff")
    light_surface_color = ColorProp("#FAFAFA")
    light_font_color = ColorProp("#030405")
    light_container_border = BorderProp((1, "solid", "#CFD0D1"))
    light_input_border = BorderProp((1, "solid", "#CFD0D1"))
    light_modal_cover_color = ColorProp("#00000018")

    # Computed props based on the mode. `light_or_dark("foo")` assumes
    # that both `light_foo` and `dark_foo` are defined props on theme.

    # TODO: These should be read-only
    background_color = ColorProp(light_or_dark("background_color"))
    surface_color = ColorProp(light_or_dark("surface_color"))
    font_color = ColorProp(light_or_dark("font_color"))
    container_border = BorderProp(light_or_dark("container_border"))
    input_border = BorderProp(light_or_dark("input_border"))
    modal_cover_color = ColorProp(light_or_dark("modal_cover_color"))

    @property
    def is_dark(self):
        return self.mode == "dark" or (
            self.mode == "system" and self.system_mode == "dark"
        )


def theme_prop(propname):
    """Gotcha: All theme prop classes must be constructable with a single
    value argument for this function to work generally. E.g., if
    `theme` had a prop with two args like:

      foo = OptionProp("x", ("x", "y"))

    then `theme_prop('foo')` would raise an exception.
    """
    existing_prop = getattr(theme, propname)
    return existing_prop.__class__(lambda e: getattr(theme(), propname))
